#!/usr/bin/env python3

# std
from pathlib import Path, PurePath
import sqlite3
from typing import Optional, Dict, Union
import time

# ours
import ankipandas.paths
import ankipandas.raw as raw
from ankipandas.ankidf import AnkiDataFrame
from ankipandas.util.log import log


class Collection(object):
    def __init__(self, path=None, user=None):
        """ Initialize :class:`~ankipandas.collection.Collection` object.

        Args:
            path: (Search) path to database see
                :py:func:`~ankipandas.paths.db_path_input` for more
                information.
            user: Anki user name. See
                :py:func:`~ankipandas.paths.db_path_input` for more
                information.

        Examples:

        .. code-block:: python

            from ankipandas import Collection

            # Let ankipandas find the db for you
            col = Collection()

            # Let ankipandas find the db for this user (important if you have
            # more than one user account in Anki)
            col = Collection(user="User 1")

            # Specify full path to Anki's database
            col = Collection("/full/path/to/collection.anki2")

            # Speficy partial path to Anki's database and specify user
            col = Collection("/partial/path/to/collection", user="User 1")

        """
        path = ankipandas.paths.db_path_input(path, user=user)

        #: Path to currently loaded database
        self.path = path  # type: Path

        log.info("Loaded db from {}".format(self.path))

        #: Opened Anki database (:class:`sqlite3.Connection`)
        self.db = raw.load_db(self.path)  # type: sqlite3.Connection

        #: Should be accessed with _get_item!
        self.__items = {
            "notes": None,
            "cards": None,
            "revs": None,
        }  # type: Dict[str, AnkiDataFrame]

        #: Should be accessed with _get_original_item!
        self.__original_items = {
            "notes": None,
            "cards": None,
            "revs": None,
        }  # type: Dict[str, AnkiDataFrame]

    def _get_original_item(self, item):
        r = self.__original_items[item]
        if r is None:
            if item in ["notes", "revs", "cards"]:
                r = AnkiDataFrame.init_with_table(self, item)
                self.__original_items[item] = r
        return r

    def _get_item(self, item):
        r = self.__items[item]
        if r is None:
            r = self._get_original_item(item).copy(True)
            self.__items[item] = r
        return r

    @property
    def notes(self) -> AnkiDataFrame:
        """ Notes as :class:`ankipandas.ankidf.AnkiDataFrame`. """
        return self._get_item("notes")

    @notes.setter
    def notes(self, value):
        self.__items["notes"] = value

    @property
    def cards(self) -> AnkiDataFrame:
        """ Cards as :class:`ankipandas.ankidf.AnkiDataFrame`. """
        return self._get_item("cards")

    @cards.setter
    def cards(self, value):
        self.__items["cards"] = value

    @property
    def revs(self) -> AnkiDataFrame:
        """ Reviews as :class:`ankipandas.ankidf.AnkiDataFrame`. """
        return self._get_item("revs")

    @revs.setter
    def revs(self, value):
        self.__items["revs"] = value

    def empty_notes(self):
        """ Similar :class:`ankipandas.ankidf.AnkiDataFrame`
        to :attr:`notes`, but without any rows. """
        return AnkiDataFrame.init_with_table(self, "notes", empty=True)

    def empty_cards(self):
        """ Similar :class:`ankipandas.ankidf.AnkiDataFrame`
        to :attr:`cards`, but without any rows. """
        return AnkiDataFrame.init_with_table(self, "cards", empty=True)

    def empty_revs(self):
        """ Similar :class:`ankipandas.ankidf.AnkiDataFrame`
        to :attr:`revs`, but without any rows. """
        return AnkiDataFrame.init_with_table(self, "revs", empty=True)

    def summarize_changes(self, output="print") -> Optional[Dict[str, dict]]:
        """ Summarize changes that were made with respect to the table
        as loaded from the database.
        If notes/cards/etc. were not loaded at all (and hence also definitely
        not modified), they do not appear in the output.

        Args:
            output: Output mode: 'print' (default: print)
                or 'dict' (return as dictionary of dictionaries of format
                ``{<type (cards/notes/...)>: {<key>: <value>}}``.

        Returns:
            None or dictionary of dictionaries
        """
        if output == "dict":
            as_dict = {}
            for key, value in self.__items.items():
                if value is not None:
                    as_dict[key] = value.summarize_changes(output="dict")
            return as_dict
        elif output == "print":
            for key, value in self.__items.items():
                if value is not None:
                    print("======== {} ========".format(key))
                    value.summarize_changes()
        else:
            raise ValueError("Invalid output setting: {}".format(output))

    def write(
        self,
        modify=False,
        add=False,
        delete=False,
        backup_folder: Union[PurePath, str] = None,
    ):
        """ Creates a backup of the database and then writes back the new
        data.

        .. danger::

            The switches ``modify``, ``add`` and ``delete`` will run additional
            cross-checks, but do not rely on them to 100%!

        .. warning::

            It is recommended to run :meth:`summarize_changes` before to check
            whether the changes match your expectation.

        .. note::

            Please make sure to thoroughly check your collection in Anki after
            every write process!

        Args:
            modify: Allow modification of existing items (notes, cards, etc.)
            add: Allow adding of new items (notes, cards, etc.)
            delete: Allow deletion of items (notes, cards, etc.)
            backup_folder: Path to backup folder. If None is given, the backup
                is created in the Anki backup directory (if found).

        Returns:
            None
        """
        if not modify and not add and not delete:
            log.warning(
                "Please set modify=True, add=True or delete=True, you're"
                "literally not allowing me any modification at all."
            )
            return None
        backup_path = ankipandas.paths.backup_db(
            self.path, backup_folder=backup_folder
        )
        log.info("Backup created at {}.".format(backup_path.resolve()))
        for key, value in self.__items.items():
            if value is None:
                log.debug("Write: Skipping {}, because it's None.".format(key))
                continue
            if key in ["notes", "cards", "revs"]:
                if not delete:
                    ndeleted = sum(value.was_deleted())
                    if ndeleted:
                        raise ValueError(
                            "You specified delete=False, but {} rows of item "
                            "{} would be deleted.".format(ndeleted, key)
                        )
                if not modify:
                    nmodified = sum(value.was_modified(na=False))
                    if nmodified:
                        raise ValueError(
                            "You specified modify=False, but {} rows of item "
                            "{} would be modified.".format(nmodified, key)
                        )
                if not add:
                    nadded = sum(value.was_added())
                    if nadded:
                        raise ValueError(
                            "You specified add=False, but {} rows of item "
                            "{} would be modified.".format(nadded, key)
                        )

                mode = "replace"
                if modify and not add and not delete:
                    mode = "update"
                if add and not modify and not delete:
                    mode = "append"
                raw.set_table(self.db, value.raw(), table=key, mode=mode)
        info = raw.get_info(self.db)
        info["mod"] = int(time.time() * 1000)  # Modification time stamp
        info["usn"] = -1  # Signals update needed
        if self.__items["notes"] is not None:
            missing_tags = list(
                set(info["tags"].keys())
                - set(self.__items["notes"].list_tags())
            )
            for tag in missing_tags:
                # I'm assuming that this is the usn (update sequence number)
                # of the tags
                info["tags"][tag] = -1



#!/usr/bin/env python3

# std
import unittest

# ours
from ankipandas.util.types import *


class TestTypes(unittest.TestCase):
    def setUp(self):
        pass

    def test_is_list_like(self):
        self.assertTrue(is_list_like([1, 2]))
        self.assertTrue(is_list_like((1, 2)))
        self.assertFalse(is_list_like("asdf"))

    def test_is_list_list_like(self):
        self.assertTrue(is_list_list_like([[1, 2]]))
        self.assertTrue(is_list_list_like([(1, 2)]))
        self.assertFalse(is_list_list_like([(1, 2), 3]))

    def test_is_list_dict_like(self):
        self.assertTrue(is_list_dict_like([{1: 3}, {4: 5}]))
        self.assertTrue(is_list_dict_like([]))
        self.assertFalse(is_list_dict_like([(1, 2), (4, 5)]))

    def test_is_dict_list_like(self):
        self.assertTrue(is_dict_list_like({1: [], 2: (3, 4)}))
        self.assertTrue(is_dict_list_like({}))
        self.assertFalse(is_dict_list_like([(1, 2), (4, 5)]))


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import ankipandas.raw
import ankipandas.paths
import ankipandas.util
import ankipandas.paths
import ankipandas.collection
from ankipandas.collection import Collection
from ankipandas.paths import find_db, db_path_input
from ankipandas.ankidf import AnkiDataFrame
from ankipandas.util.log import log, set_log_level



#!/usr/bin/env python3

# std
import collections
import unittest
import tempfile
from typing import List, Any

# 3rd
from randomfiletree import sample_random_elements, iterative_gaussian_tree

# ours
import ankipandas.paths as paths
from ankipandas._columns import *
from ankipandas.util.misc import flatten_list_list


def touch_file_in_random_folders(basedir, filename: str, n=1) -> List[Path]:
    """ Create files in random folders.

    Args:
        basedir: Starting directory
        filename: Filename of the files to create
        n: Number of files to create

    Returns:
        List of files that were created.
    """
    files = set()
    for d in sample_random_elements(
        basedir, n_dirs=n, n_files=0, onfail="ignore"
    )[0]:
        p = Path(d) / filename
        p.touch()
        files.add(p)
    return list(files)


class TestFindDatabase(unittest.TestCase):
    def setUp(self):
        self.dirs = {
            "nothing": tempfile.TemporaryDirectory(),
            "multiple": tempfile.TemporaryDirectory(),
            "perfect": tempfile.TemporaryDirectory(),
        }
        for d in self.dirs.values():
            iterative_gaussian_tree(
                d.name,
                repeat=5,
                nfolders=1,
                min_folders=1,
                nfiles=2,
                min_files=1,
            )
        self.dbs = {
            "nothing": [],
            "multiple": touch_file_in_random_folders(
                self.dirs["multiple"].name, "collection.anki2", 10
            ),
            "perfect": touch_file_in_random_folders(
                self.dirs["perfect"].name, "collection.anki2", 1
            ),
        }
        self.maxDiff = None

    def test_db_path_input_nexist(self):
        with self.assertRaises(FileNotFoundError):
            paths.db_path_input("/x/y/z")

    def test_db_path_input_multiple(self):
        with self.assertRaises(ValueError):
            paths.db_path_input(self.dirs["multiple"].name)

    def test_db_path_input_nothing(self):
        with self.assertRaises(ValueError):
            paths.db_path_input(self.dirs["nothing"].name)

    def test_db_path_input_perfect(self):
        self.assertEqual(
            paths.db_path_input(self.dirs["perfect"].name),
            self.dbs["perfect"][0],
        )

    def test__find_database(self):
        for d in self.dirs:
            a = sorted(
                map(
                    str,
                    flatten_list_list(
                        paths._find_db(
                            self.dirs[d].name,
                            maxdepth=None,
                            break_on_first=False,
                        ).values()
                    ),
                )
            )
            b = sorted(map(str, self.dbs[d]))
            self.assertListEqual(a, b)

    def test__find_database_filename(self):
        # If doesn't exist
        self.assertEqual(
            paths._find_db(
                Path("abc/myfilename.txt"), filename="myfilename.txt"
            ),
            {},
        )
        tmpdir = tempfile.TemporaryDirectory()
        dir_path = Path(tmpdir.name) / "myfolder"
        file_path = dir_path / "myfilename.txt"
        dir_path.mkdir()
        file_path.touch()
        self.assertEqual(
            paths._find_db(file_path, filename="myfilename.txt"),
            collections.defaultdict(list, {"myfolder": [file_path]}),
        )
        tmpdir.cleanup()

    def test_find_database(self):
        with self.assertRaises(ValueError):
            paths.find_db(self.dirs["nothing"].name, break_on_first=False)
        with self.assertRaises(ValueError):
            paths.find_db(self.dirs["multiple"].name, break_on_first=False)
            print(self.dbs["multiple"])
        self.assertEqual(
            str(paths.find_db(self.dirs["perfect"].name, break_on_first=False)),
            str(self.dbs["perfect"][0]),
        )

    def tearDown(self):
        for d in self.dirs.values():
            d.cleanup()


class TestBackup(unittest.TestCase):
    def setUp(self):
        self.tmpdir = tempfile.TemporaryDirectory()
        self.tmpdir_path = Path(self.tmpdir.name)
        (self.tmpdir_path / "collection.anki2").touch()
        (self.tmpdir_path / "backups").mkdir()

        self.tmpdir_only_db = tempfile.TemporaryDirectory()
        self.tmpdir_only_db_path = Path(self.tmpdir_only_db.name)
        (self.tmpdir_only_db_path / "collection.anki2").touch()

    def tearDown(self):
        self.tmpdir.cleanup()
        self.tmpdir_only_db.cleanup()

    def test_get_anki_backup_folder(self):
        self.assertEqual(
            str(
                paths.get_anki_backup_folder(
                    self.tmpdir_path / "collection.anki2"
                )
            ),
            str(self.tmpdir_path / "backups"),
        )

    def test_get_anki_backup_folder_raise(self):
        with self.assertRaises(FileNotFoundError):
            paths.get_anki_backup_folder(self.tmpdir_path / "asdf")
        with self.assertRaises(ValueError):
            paths.get_anki_backup_folder(
                self.tmpdir_only_db_path / "collection.anki2"
            )
        paths.get_anki_backup_folder(
            self.tmpdir_only_db_path / "collection.anki2", nexist="ignore"
        )

    def test_backup_db_auto(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            db_path = Path(tmpdir) / "collection.anki2"
            db_path.touch()
            backup_folder = db_path.parent / "backups"
            backup_folder.mkdir()
            backup_path = paths.backup_db(db_path)
            self.assertTrue(backup_path.is_file())
            self.assertTrue(backup_path.parent == backup_folder)

    def test_backup_db_custom(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            db_path = Path(tmpdir) / "collection.anki2"
            db_path.touch()
            backup_folder = db_path.parent / "myfolder"
            backup_path = paths.backup_db(db_path, backup_folder=backup_folder)
            self.assertTrue(backup_path.is_file())
            self.assertTrue(backup_path.parent == backup_folder)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

import re
from hashlib import sha1
from html.entities import name2codepoint

# Implementation directly copied from Anki (anki/anki/utils.py).
# Only a bit of PEP8ing and making things private.

_reComment = re.compile("(?s)<!--.*?-->")
_reStyle = re.compile("(?si)<style.*?>.*?</style>")
_reScript = re.compile("(?si)<script.*?>.*?</script>")
_reTag = re.compile("(?s)<.*?>")
_reEnts = re.compile(r"&#?\w+;")
_reMedia = re.compile("(?i)<img[^>]+src=[\"']?([^\"'>]+)[\"']?[^>]*>")


def _strip_html(s):
    s = _reComment.sub("", s)
    s = _reStyle.sub("", s)
    s = _reScript.sub("", s)
    s = _reTag.sub("", s)
    s = _ents_to_txt(s)
    return s


def _strip_html_media(s):
    """ Strip HTML but keep media filenames """
    s = _reMedia.sub(" \\1 ", s)
    return _strip_html(s)


def _ents_to_txt(html):
    # entitydefs defines nbsp as \xa0 instead of a standard space, so we
    # replace it first
    html = html.replace("&nbsp;", " ")

    def fixup(m):
        text = m.group(0)
        if text[:2] == "&#":
            # character reference
            try:
                if text[:3] == "&#x":
                    return chr(int(text[3:-1], 16))
                else:
                    return chr(int(text[2:-1]))
            except ValueError:
                pass
        else:
            # named entity
            try:
                text = chr(name2codepoint[text[1:-1]])
            except KeyError:
                pass
        return text  # leave as is

    return _reEnts.sub(fixup, html)


def _checksum(data):
    if isinstance(data, str):
        data = data.encode("utf-8")
    return sha1(data).hexdigest()


def field_checksum(data: str) -> int:
    """ 32 bit unsigned number from first 8 digits of sha1 hash.
    Apply this to the first field to the the field checksum that is used by
    Anki to detect duplicates.

    Args:
        data: string like

    Returns:
        int
    """
    return int(_checksum(_strip_html_media(data).encode("utf-8"))[:8], 16)



#!/usr/bin/env python3

""" DataFrame utilities. """

# 3rd
import pandas as pd


def replace_df_inplace(df: pd.DataFrame, df_new: pd.DataFrame) -> None:
    """ Replace dataframe 'in place'.

    Args:
        df: :class:`pandas.DataFrame` to be replaced
        df_new: :class:`pandas.DataFrame` to replace the previous one

    Returns:
        None
    """
    # Drop all ROWs (not columns)
    if df.index.any():
        df.drop(df.index, inplace=True)
    for col in df_new.columns:
        df[col] = df_new[col]
    drop_cols = set(df.columns) - set(df_new.columns)
    if drop_cols:
        df.drop(drop_cols, axis=1, inplace=True)


# todo: this might be made more elegant in the future for sure...
# fixme: This removes items whenever it can't merge!
def merge_dfs(
    df: pd.DataFrame,
    df_add: pd.DataFrame,
    id_df: str,
    inplace=False,
    id_add="id",
    prepend="",
    replace=False,
    prepend_clash_only=True,
    columns=None,
    drop_columns=None,
):
    """
    Merge information from two dataframes.

    Args:
        df: Original :class:`pandas.DataFrame`
        df_add: :class:`pandas.DataFrame` to be merged with original
            :class:`pandas.DataFrame`
        id_df: Column of original dataframe that contains the id along which
            we merge.
        inplace: If False, return new dataframe, else update old one
        id_add: Column of the new dataframe that contains the id along which
            we merge
        prepend: Prepend a string to the column names from the new dataframe
        replace: Replace columns
        prepend_clash_only: Only prepend string to the column names from the
            new dataframe if there is a name clash.
        columns: Keep only these columns
        drop_columns: Drop these columns

    Returns:
        New merged :class:`pandas.DataFrame`
    """
    # Careful: Do not drop the id column until later (else we can't merge)
    # Still, we want to remove as much as possible here, because it's probably
    # better performing
    if columns:
        df_add = df_add.drop(
            set(df_add.columns) - (set(columns) | {id_add}), axis=1
        )
    if drop_columns:
        df_add = df_add.drop(set(drop_columns) - {id_add}, axis=1)
    # Careful: Rename columns after dropping unwanted ones
    if prepend_clash_only:
        col_clash = set(df.columns) & set(df_add.columns)
        rename_dict = {col: prepend + col for col in col_clash}
    else:
        rename_dict = {col: prepend + col for col in df_add.columns}
    df_add = df_add.rename(columns=rename_dict)
    # Careful: Might have renamed id_add as well
    if id_add in rename_dict:
        id_add = rename_dict[id_add]

    if replace:
        # Simply remove all potential clashes
        replaced_columns = set(df_add.columns).intersection(set(df.columns))
        df = df.drop(replaced_columns, axis=1)

    merge_kwargs = {}

    if id_add in df_add.columns:
        merge_kwargs["right_on"] = id_add
    elif id_add == df_add.index.name:
        merge_kwargs["right_index"] = True
    else:
        raise ValueError("'{}' is neither index nor column.".format(id_add))

    if id_df in df.columns:
        merge_kwargs["left_on"] = id_df
    elif id_df == df.index.name:
        merge_kwargs["left_index"] = True
    else:
        raise ValueError("'{}' is neither index nor column.".format(id_df))

    df_merge = df.merge(df_add, **merge_kwargs)

    # Now remove id_add if it was to be removed
    # Careful: 'in' doesn't work with None
    if (columns and id_add not in columns) or (
        drop_columns and id_add in drop_columns
    ):
        df_merge.drop(id_add, axis=1, inplace=True)

    # Make sure we don't have two ID columns
    new_id_add_col = id_add
    if id_add in rename_dict:
        new_id_add_col = rename_dict[id_add]
    if new_id_add_col in df_merge.columns and id_df != new_id_add_col:
        df_merge.drop(new_id_add_col, axis=1, inplace=True)

    if inplace:
        return replace_df_inplace(df, df_merge)
    else:
        return df_merge



#!/usr/bin/env python3

# std
import unittest

# 3rd
import pandas as pd

# ours
from ankipandas.util.dataframe import replace_df_inplace


class TestUtils(unittest.TestCase):
    def test__replace_df_inplace(self):
        df = pd.DataFrame({"a": [1, 2], "b": [3, 4]})
        df_new = pd.DataFrame({"a": [1]})
        replace_df_inplace(df, df_new)
        self.assertEqual(len(df), 1)
        self.assertEqual(len(df.columns), 1)
        self.assertListEqual(list(df["a"].values), [1])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

# std
import copy
import time

# 3rd
import numpy as np
import pandas as pd
import pathlib
from typing import Union, List, Dict, Optional

# ours
import ankipandas.paths
import ankipandas.raw as raw
import ankipandas.util.dataframe
from ankipandas.util.dataframe import replace_df_inplace
import ankipandas._columns as _columns
from ankipandas.util.misc import invert_dict, flatten_list_list
from ankipandas.util.log import log
from ankipandas.util.checksum import field_checksum
from ankipandas.util.guid import guid as generate_guid
from ankipandas.util.types import *


class AnkiDataFrame(pd.DataFrame):
    #: Additional attributes of a :class:`AnkiDataFrame` that a normal
    #: :class:`pandas.DataFrame` does not posess. These will be copied in the
    #: constructor.
    _attributes = (
        "col",
        "_anki_table",
        "fields_as_columns_prefix",
        "_fields_format",
        "_df_format",
    )

    def __init__(self, *args, **kwargs):
        """ Initializes a blank :class:`AnkiDataFrame`.

        .. warning::

            It is recommended to directly initialize this class with the notes,
            cards or revs table, using one of the methods
            :meth:`.notes`, :meth:`.cards` or :meth:`.revs` instead!

        Args:
            *args: Internal use only. See arguments of
                :class:`pandas.DataFrame`.
            **kwargs: Internal use only. See arguments of
                :class:`pandas.DataFrame`.
        """
        super().__init__(*args, **kwargs)

        # IMPORTANT: Make sure to add all attributes to the class variable
        # :attr:`._attributes`. Also all of them have to be initialized as None!
        # (see the code where we copy attributes).

        # todo: document
        self.col = None

        #: Type of anki table: 'notes', 'cards' or 'revlog'. This corresponds to
        #: the meaning of the ID row.
        self._anki_table = None  # type: str

        #: Prefix for fields as columns. Default is ``nfld_``.
        self.fields_as_columns_prefix = "nfld_"

        #: Fields format: ``none``, ``list`` or ``columns`` or ``in_progress``,
        #:   or ``anki`` (default)
        self._fields_format = "anki"

        #: Overal structure of the dataframe ``anki``, ``ours``, ``in_progress``
        self._df_format = None  # type: str

        # todo: is this serving any purpose? Coverage shows it never runs.
        if len(args) == 1 and isinstance(args[0], AnkiDataFrame):
            self._copy_attrs_from(args[0])

    @property
    def _constructor(self):
        """ This needs to be overriden so that any DataFrame operations do not
        return a :class:`pandas.DataFrame` but a :class:`AnkiDataFrame`."""

        def __constructor(*args, **kw):
            df = self.__class__(*args, **kw)
            self._copy_attrs_to(df)
            return df

        return __constructor

    def _copy_attrs_to(self, df):
        """ Copy all additional attributes of this class to another instance.
        Also see :attr:`self._attributes`.
        """
        for attr in self._attributes:
            df.__dict__[attr] = getattr(self, attr, None)

    def _copy_attrs_from(self, df):
        """ Copy all additional attributes of this class from another instance.
        Also see :attr:`self._attributes`.
        """
        for attr in self._attributes:
            self.__dict__[attr] = getattr(df, attr, None)

    # Constructors
    # ==========================================================================

    def _get_table(self, col, table, empty=False):
        self._anki_table = table
        self._df_format = "anki"
        self.col = col

        if empty:
            df = raw.get_empty_table(table)
        else:
            df = raw.get_table(col.db, table)

        replace_df_inplace(self, df)
        self.normalize(inplace=True)

    @classmethod
    def init_with_table(cls, col, table, empty=False):
        new = AnkiDataFrame()
        new._get_table(col, table, empty=empty)
        return new

    # Fixes
    # ==========================================================================

    def equals(self, other):
        return pd.DataFrame(self).equals(other)

    def append(self, *args, **kwargs):
        ret = pd.DataFrame.append(self, *args, **kwargs)
        self._copy_attrs_to(ret)
        ret.astype(_columns.dtype_casts2[self._anki_table])
        return ret

    def update(self, *args, **kwargs):
        super(AnkiDataFrame, self).update(*args, **kwargs)
        # Fix https://github.com/pandas-dev/pandas/issues/4094
        for col, typ in _columns.dtype_casts2[self._anki_table].items():
            self[col] = self[col].astype(typ)

    # ==========================================================================

    def _invalid_table(self):
        raise ValueError("Invalid table: {}.".format(self._anki_table))

    def _check_df_format(self):
        if self._df_format == "in_progress":
            raise ValueError(
                "Previous call to normalize() or raw() did not terminate "
                "succesfully. This is usually a very bad sign, but you can "
                "try calling them again with the force option: raw(force=True) "
                "or raw(force=True) and see if that works."
            )
        elif self._df_format == "anki":
            pass
        elif self._df_format == "ours":
            pass
        else:
            raise ValueError(
                "Unknown value of _df_format: {}".format(self._df_format)
            )

    def _check_our_format(self):
        self._check_df_format()
        if not self._df_format == "ours":
            raise ValueError(
                "This operation is not supported for AnkiDataFrames in the "
                "'raw' format. Perhaps you called raw() before or used the "
                "raw=True option when loading? You can try switching to the "
                "required format using the normalize() method."
            )

    # Properties
    # ==========================================================================

    @property
    def db(self):
        """Opened Anki database (:class:`sqlite3.Connection`)"""
        return self.col.db

    # IDs
    # ==========================================================================

    @property
    def id(self):
        """ Return note/card/review ID as :class:`pandas.Series` of integers.
        """
        if self._anki_table == "notes":
            return self.nid
        elif self._anki_table == "cards":
            return self.cid
        elif self._anki_table == "revs":
            return self.rid
        else:
            self._invalid_table()

    @property
    def nid(self):
        """ Note ID as :class:`pandas.Series` of integers. """
        if self._anki_table == "notes":
            return self.index
        elif self._anki_table == "cards":
            if "nid" not in self.columns:
                raise ValueError(
                    "You seem to have removed the 'nid' column. That was not "
                    "a good idea. Cannot get note ID anymore."
                )
            else:
                return self["nid"]
        elif self._anki_table == "revs":
            if "nid" in self.columns:
                return self["nid"]
            else:
                return self.cid.map(raw.get_cid2nid(self.db))
        else:
            self._invalid_table()

    @nid.setter
    def nid(self, value):
        if self._anki_table == "notes":
            raise ValueError(
                "Note ID column should already be index and notes.nid() will "
                "always return this index. Therefore you should not set nid "
                "to a column."
            )
        else:
            self["nid"] = value

    @property
    def cid(self):
        """ Card ID as :class:`pandas.Series` of integers. """
        if self._anki_table == "cards":
            return self.index
        if self._anki_table == "revs":
            if "cid" not in self.columns:
                raise ValueError(
                    "You seem to have removed the 'cid' column. That was not "
                    "a good idea. Cannot get card ID anymore."
                )
            else:
                return self["cid"]
        elif self._anki_table == "notes":
            raise ValueError(
                "Notes can belong to multiple cards. Therefore it is impossible"
                " to associate a card ID with them."
            )
        else:
            self._invalid_table()

    @cid.setter
    def cid(self, value):
        if self._anki_table == "cards":
            raise ValueError(
                "Card ID column should already be index and notes.cid() will "
                "always return this index. Therefore you should not set cid "
                "to a column."
            )
        elif self._anki_table == "revs":
            self["cid"] = value
        else:
            raise ValueError(
                "Notes can belong to multiple cards. Therefore please "
                " do not associate a card ID with them."
            )

    @property
    def rid(self):
        """ Review ID as :class:`pandas.Series` of integers. """
        if self._anki_table == "revs":
            return self.index
        else:
            if "rid" in self.columns:
                return self["rid"]
            else:
                raise ValueError(
                    "Review index is only available for the 'revs' table by"
                    " default."
                )

    # noinspection PyUnusedLocal
    @rid.setter
    def rid(self, value):
        if self._anki_table == "revs":
            raise ValueError(
                "Review ID column should already be index and notes.rid() will "
                "always return this index. Therefore you should not set rid "
                "to a column."
            )
        else:
            raise ValueError(
                "Setting a review index 'rid' makes no sense in "
                "tables other than 'rev'."
            )

    @property
    def mid(self):
        """ Model ID as :class:`pandas.Series` of integers. """
        if self._anki_table in ["notes"]:
            if "nmodel" not in self.columns:
                raise ValueError(
                    "You seem to have removed the 'nmodel' column. That was not"
                    " a good idea. Cannot get model ID anymore."
                )
            else:
                return self["nmodel"].map(raw.get_model2mid(self.db))
        if self._anki_table in ["revs", "cards"]:
            if "nmodel" in self.columns:
                return self["nmodel"].map(raw.get_model2mid(self.db))
            else:
                return self.nid.map(raw.get_nid2mid(self.db))
        else:
            self._invalid_table()

    @mid.setter
    def mid(self, value):
        if self._anki_table == "notes":
            log.warning(
                "You can set an additional 'mid' column, but this will always"
                " be overwritten with the information from the 'nmodel' "
                "column."
            )
        self["mid"] = value

    @property
    def did(self):
        """ Deck ID as :class:`pandas.Series` of integers. """
        if self._anki_table == "cards":
            if "cdeck" not in self.columns:
                raise ValueError(
                    "You seem to have removed the 'cdeck' column. That was not "
                    "a good idea. Cannot get deck ID anymore."
                )
            return self["cdeck"].map(raw.get_deck2did(self.db))
        elif self._anki_table == "notes":
            raise ValueError(
                "Notes can belong to multiple decks. Therefore it is impossible"
                " to associate a deck ID with them."
            )
        elif self._anki_table == "revs":
            return self.cid.map(raw.get_cid2did(self.db))
        else:
            self._invalid_table()

    @did.setter
    def did(self, value):
        if self._anki_table == "cards":
            log.warning(
                "You can set an additional deck ID 'did' column, but this "
                "will always be overwritten with the information from the "
                "'cdeck' column."
            )
        self["did"] = value

    @property
    def odid(self):
        """ Original deck ID for cards in filtered deck as
        :class:`pandas.Series` of integers.
        """
        if self._anki_table == "cards":
            if "odeck" not in self.columns:
                raise ValueError(
                    "You seem to have removed the 'odeck' column. That was not "
                    "a good idea. Cannot get original deck ID anymore."
                )
            return self["odeck"].map(raw.get_deck2did(self.db))
        elif self._anki_table == "revs":
            if "odeck" in self.columns:
                return self["odeck"].map(raw.get_deck2did(self.db))
        elif self._anki_table == "notes":
            raise ValueError(
                "The original deck ID (odid) is not availabale for the notes "
                "table."
            )
        else:
            self._invalid_table()

    @odid.setter
    def odid(self, value):
        if self._anki_table == "cards":
            log.warning(
                "You can set an additional 'odid' column, but this will always"
                " be overwritten with the information from the 'odeck' "
                "column."
            )
        self["odid"] = value

    # Merge tables
    # ==========================================================================

    def merge_notes(
        self,
        inplace=False,
        columns=None,
        drop_columns=None,
        prepend="n",
        prepend_clash_only=True,
    ):
        """ Merge note table into existing dataframe.

        Args:
            inplace: If False, return new dataframe, else update old one
            columns: Columns to merge
            drop_columns: Columns to ignore when merging
            prepend: Prepend this string to fields from note table
            prepend_clash_only: Only prepend the ``prepend`` string when column
                names would otherwise clash.

        Returns:
            New :class:`AnkiDataFrame` if inplace==True, else None
        """
        self._check_our_format()
        if self._anki_table == "notes":
            raise ValueError(
                "AnkiDataFrame was already initialized as a table of type"
                " notes, therefore merge_notes() doesn't make any sense."
            )
        elif self._anki_table == "revs":
            self["nid"] = self.nid
        return ankipandas.util.dataframe.merge_dfs(
            df=self,
            df_add=self.col.notes,
            id_df="nid",
            id_add="nid",
            inplace=inplace,
            prepend=prepend,
            prepend_clash_only=prepend_clash_only,
            columns=columns,
            drop_columns=drop_columns,
        )

    def merge_cards(
        self,
        inplace=False,
        columns=None,
        drop_columns=None,
        prepend="c",
        prepend_clash_only=True,
    ):
        """
        Merges information from the card table into the current dataframe.

        Args:
            inplace: If False, return new dataframe, else update old one
            columns:  Columns to merge
            drop_columns:  Columns to ignore when merging
            prepend: Prepend this string to fields from card table
            prepend_clash_only: Only prepend the ``prepend`` string when column
                names would otherwise clash.

        Returns:
            New :class:`AnkiDataFrame` if inplace==True, else None
        """
        if self._anki_table == "cards":
            raise ValueError(
                "AnkiDataFrame was already initialized as a table of type"
                " cards, therefore merge_cards() doesn't make any sense."
            )
        elif self._anki_table == "notes":
            raise ValueError(
                "One note can correspond to more than one card, therefore it "
                "it is not supported to merge the cards table into the "
                "notes table."
            )
        self._check_our_format()
        return ankipandas.util.dataframe.merge_dfs(
            df=self,
            df_add=self.col.cards,
            id_df="cid",
            inplace=inplace,
            columns=columns,
            drop_columns=drop_columns,
            id_add="cid",
            prepend=prepend,
            prepend_clash_only=prepend_clash_only,
        )

    # Toggle format
    # ==========================================================================

    def fields_as_columns(self, inplace=False, force=False):
        """
        In the 'notes' table, the field contents of the notes is contained in
        one column ('flds') by default. With this method, this column can be
        split up into a new column for every field.

        Args:
            inplace: If False, return new dataframe, else update old one
            force: Internal use

        Returns:
            New :class:`pandas.DataFrame` if inplace==True, else None
        """
        if not force:
            self._check_our_format()
        if not inplace:
            df = self.copy(True)
            df.fields_as_columns(inplace=True)
            return df

        if "nflds" not in self.columns:
            raise ValueError("Could not find fields column 'nflds'.")

        if self._fields_format == "columns":
            log.warning(
                "Fields are already as columns."
                " Returning without doing anything."
            )
            return
        elif self._fields_format == "in_progress" and not force:
            raise ValueError(
                "It looks like the last call to fields_as_list or"
                "fields_as_columns was not successful, so you better start "
                "over."
            )
        elif self._fields_format == "list":
            pass
        else:
            raise ValueError(
                "Unknown _fields_format: {}".format(self._fields_format)
            )

        self._fields_format = "in_progress"
        # fixme: What if one field column is one that is already in use?
        prefix = self.fields_as_columns_prefix
        mids = self.mid.unique()
        for mid in mids:
            if mid == 0:
                continue
            df_model = self[self.mid == mid]
            fields = pd.DataFrame(df_model["nflds"].tolist())
            field_names = raw.get_mid2fields(self.db)[mid]
            for field in field_names:
                if prefix + field not in self.columns:
                    self[prefix + field] = ""
            for ifield, field in enumerate(field_names):
                # todo: can we speed this up?
                self.loc[self.mid == mid, [prefix + field]] = pd.Series(
                    fields[ifield].tolist(),
                    index=self.loc[self.mid == mid].index,
                )
        self.drop("nflds", axis=1, inplace=True)
        self._fields_format = "columns"

    def fields_as_list(self, inplace=False, force=False):
        """
        This reverts :meth:`.fields_as_columns`, all columns that represented
        field contents are now merged into one column 'nflds'.

        Args:
            inplace: If False, return new dataframe, else update old one
            force: Internal use

        Returns:
            New :class:`AnkiDataFrame` if inplace==True, else None
        """
        if not force:
            self._check_our_format()
        if not inplace:
            df = self.copy(True)
            df.fields_as_list(inplace=True)
            return df

        if self._fields_format == "list":
            log.warning(
                "Fields are already as list. Returning without doing anything."
            )
            return
        elif self._fields_format == "in_progress" and not force:
            raise ValueError(
                "It looks like the last call to fields_as_list or"
                "fields_as_columns was not successful, so you better start "
                "over."
            )
        elif self._fields_format == "columns":
            pass
        else:
            raise ValueError(
                "Unknown _fields_format: {}".format(self._fields_format)
            )

        self._fields_format = "in_progress"
        mids = self.mid.unique()
        to_drop = []
        for mid in mids:
            fields = raw.get_mid2fields(self.db)[mid]
            fields = [self.fields_as_columns_prefix + field for field in fields]
            self.loc[self.mid == mid, "nflds"] = pd.Series(
                self.loc[self.mid == mid, fields].values.tolist(),
                index=self.loc[self.mid == mid].index,
            )
            # Careful: Do not delete the fields here yet, other models
            # might still use them
            to_drop.extend(fields)
        self.drop(to_drop, axis=1, inplace=True)
        self._fields_format = "list"

    # Quick access
    # ==========================================================================

    def _check_tag_col(self):
        if "ntags" not in self.columns:
            raise ValueError(
                "Tag column 'ntags' doesn't exist. Perhaps you forgot to merge "
                "the notes into your table?"
            )

    def list_tags(self) -> List[str]:
        """ Return sorted list of all tags in the current table. """
        if "ntags" not in self.columns:
            raise ValueError(
                "Tags column 'ntags' not present. Either use the notes table"
                " or merge it into your table."
            )
        else:
            return sorted(
                list(
                    set(
                        [item for lst in self["ntags"].tolist() for item in lst]
                    )
                )
            )

    def list_decks(self) -> List[str]:
        """ Return sorted list of deck names in the current table. """
        if "cdeck" not in self.columns:
            raise ValueError(
                "Deck column 'cdeck' not present. Either use the cards table "
                "or merge it into your table."
            )
        else:
            decks = sorted(list(self["cdeck"].unique()))
            if "" in decks:
                decks.remove("")
            return decks

    def list_models(self):
        """ Return sorted list of model names in the current table. """
        if "nmodel" not in self.columns:
            raise ValueError(
                "Model column 'nmodel' not present. Either use the notes table"
                " or merge it into your table."
            )
        return sorted(list(self["nmodel"].unique()))

    def has_tag(self, tags=None):
        """ Checks whether row has a certain tag ('ntags' column).

        Args:
            tags: String or list thereof. In the latter case, True is returned
                if the row contains any of the specified tags.
                If None (default), True is returned if the row has any tag at
                all.

        Returns:
            Boolean :class:`pd.Series`

        Examples:

            .. code-block:: python

                # Get all tagged notes:
                notes[notes.has_tag()]
                # Get all untagged notes:
                notes[~notes.has_tag()]
                # Get all notes tagged Japanese:
                japanese_notes = notes[notes.has_tag("Japanese")]
                # Get all notes tagged either Japanese or Chinese:
                asian_notes = notes[notes.has_tag(["Japanese", "Chinese"])]
        """
        self._check_our_format()
        self._check_tag_col()

        if isinstance(tags, str):
            tags = [tags]

        if tags is not None:

            def _has_tag(other):
                return not set(tags).isdisjoint(other)

            return self["ntags"].apply(_has_tag)

        else:
            return self["ntags"].apply(bool)

    def has_tags(self, tags=None):
        """ Checks whether row contains at least the supplied tags.

        Args:
            tags: String or list thereof.
                If None (default), True is returned if the row has any tag at
                all.

        Returns:
            Boolean :class:`pd.Series`

        Examples:

            .. code-block:: python

                # Get all notes tagged BOTH Japanese or Chinese
                bilingual_notes = notes[notes.has_tags(["Japanese", "Chinese"])]
                # Note the difference to
                asian_notes = notes[notes.has_tag(["Japanese", "Chinese"])]
        """
        self._check_our_format()
        if tags is None:
            return self.has_tag(None)
        self._check_tag_col()
        if isinstance(tags, str):
            tags = [tags]
        _has_tags = set(tags).issubset
        return self["ntags"].apply(_has_tags)

    def add_tag(self, tags, inplace=False):
        """ Adds tag ('ntags' column).

        Args:
            tags: String or list thereof.
            inplace: If False, return new dataframe, else update old one

        Returns:
            New :class:`AnkiDataFrame` if inplace==True, else None
        """
        self._check_our_format()
        if not inplace:
            df = self.copy(True)
            df.add_tag(tags, inplace=True)
            return df

        self._check_tag_col()
        if isinstance(tags, str):
            tags = [tags]

        if len(tags) == 0:
            return

        def _add_tags(other):
            return other + sorted(list(set(tags) - set(other)))

        self["ntags"] = self["ntags"].apply(_add_tags)

    def remove_tag(self, tags, inplace=False):
        """ Removes tag ('ntags' column).

        Args:
            tags: String or list thereof. If None, all tags are removed.
            inplace: If False, return new dataframe, else update old one

        Returns:
            New :class:`AnkiDataFrame` if inplace==True, else None
        """
        self._check_our_format()
        if not inplace:
            df = self.copy(True)
            df.remove_tag(tags, inplace=True)
            return df

        self._check_tag_col()
        if isinstance(tags, str):
            tags = [tags]

        if tags is not None:

            def _remove_tags(other):
                return [tag for tag in other if tag not in tags]

            self["ntags"] = self["ntags"].apply(_remove_tags)

        else:
            self["ntags"] = self["ntags"].apply(lambda _: [])

    # Compare
    # ==========================================================================

    def was_modified(self, other: pd.DataFrame = None, na=True, _force=False):
        """ Compare with original table, show which rows have changed.

        Args:
            other: Compare with this :class:`pandas.DataFrame`.
                If None (default), use original unmodified dataframe as reloaded
                from the database.
            na: Value for new or deleted columns
            _force: internal use

        Returns:
            Boolean value for each row, showing if it was modified.
        """
        if not _force:
            self._check_our_format()

        if other is None:
            other = self.col._get_original_item(self._anki_table)

        other_nids = set(other.index)
        inters = set(self.index).intersection(other_nids)
        result = pd.Series(na, index=self.index)
        new_bools = np.any(
            other.loc[other.index.isin(inters), list(self.columns)].values
            != self[self.index.isin(inters)].values,
            axis=1,
        )
        result.loc[self.index.isin(inters)] = pd.Series(
            new_bools, index=result[self.index.isin(inters)].index
        )
        return result

    def modified_columns(
        self, other: pd.DataFrame = None, _force=False, only=True
    ):
        """ Compare with original table, show which columns in which rows
        were modified.

        Args:
            other: Compare with this :class:`pandas.DataFrame`.
                If None (default), use original unmodified dataframe as reloaded
                from the database.
            only: Only show rows where at least one column is changed.
            _force: internal use

        Returns:
            Boolean value for each row, showing if it was modified. New rows
            are considered to be modified as well.
        """
        if other is None:
            other = self.init_with_table(col=self.col, table=self._anki_table)
        cols = [c for c in self.columns if c in other.columns]
        other_nids = set(other.index)
        inters = set(self.index).intersection(other_nids)
        if only:
            inters = inters.intersection(
                self[self.was_modified(other=other, _force=_force)].index
            )
        inters = sorted(list(inters))
        return pd.DataFrame(
            self.loc[inters, cols].values != other.loc[inters, cols].values,
            index=self.loc[inters].index,
            columns=cols,
        )

    def was_added(self, other: pd.DataFrame = None, _force=False):
        """ Compare with original table, show which rows were added.

        Args:
            other: Compare with this :class:`pandas.DataFrame`.
                If None (default), use original unmodified dataframe as reloaded
                from the database.
            _force: internal use

        Returns:
            Boolean value for each row, showing if it was modified. New rows
            are considered to be modified as well.
        """
        if not _force:
            self._check_our_format()

        if other is not None:
            other_ids = set(other.index)
        else:
            other_ids = set(self.col._get_original_item(self._anki_table).id)

        new_indices = set(self.index) - other_ids
        return self.index.isin(new_indices)

    def was_deleted(self, other: pd.DataFrame = None, _force=False):
        """ Compare with original table, return deleted indizes.

        Args:
            other: Compare with this :class:`pandas.DataFrame`.
                If None (default), use original unmodified dataframe as reloaded
                from the database.
            _force: internal use

        Returns:
            Sorted list of indizes.
        """
        if not _force:
            self._check_our_format()

        if other is not None:
            other_ids = set(other.index)
        else:
            other_ids = set(self.col._get_original_item(self._anki_table).id)

        deleted_indices = other_ids - set(self.index)
        return sorted(list(deleted_indices))

    # Update modification stamps and similar
    # ==========================================================================

    def _set_usn(self):
        """ Update usn (update sequence number) for all changed rows. """
        self.loc[
            self.was_modified(na=True, _force=True),
            _columns.columns_anki2ours[self._anki_table]["usn"],
        ] = -1

    def _set_mod(self):
        """ Update modification timestamps for all changed rows. """
        if self._anki_table in ["cards", "notes"]:
            self.loc[
                self.was_modified(na=True, _force=True),
                _columns.columns_anki2ours[self._anki_table]["mod"],
            ] = int(time.time())

    # todo: test
    def _set_guid(self):
        """ Update globally unique id """
        if self._anki_table == "notes":
            self.loc[~self["nguid"].apply(bool)].apply(generate_guid)

    # Raw and normalized
    # ==========================================================================

    def normalize(self, inplace=False, force=False):
        """ Bring a :class:`AnkiDataFrame` from the ``raw`` format (i.e. the
        exact format that Anki uses in its internal representation) to our
        convenient format.

        Args:
            inplace: If False, return new dataframe, else update old one
            force: If a previous conversion fails, :meth:`normalize` will
                refuse to attempt another one by default. Use this option
                to force it to attempt in anyway.

        Returns:
            New :class:`AnkiDataFrame` if inplace==True, else None
        """
        if not inplace:
            df = self.copy(True)
            df.normalize(inplace=True, force=force)
            return df

        if not force:
            self._check_df_format()
            if self._df_format == "ours":
                log.warning(
                    "Dataframe already is in our format. "
                    "Returning without doing anything."
                )
                return

        table = self._anki_table
        if table not in ["cards", "revs", "notes"]:
            self._invalid_table()

        self._df_format = "in_progress"

        # Dtypes
        # ------

        for column, typ in _columns.dtype_casts[table].items():
            self[column] = self[column].astype(typ)

        # Renames
        # -------

        self.rename(columns=_columns.columns_anki2ours[table], inplace=True)

        # Value maps
        # ----------
        # We sometimes interpret cryptic numeric values

        if table in _columns.value_maps:
            for column in _columns.value_maps[table]:
                self[column] = self[column].map(
                    _columns.value_maps[table][column]
                )

        # IDs
        # ---

        self.set_index(_columns.table2index[table], inplace=True)

        if table == "cards":
            self["cdeck"] = self["did"].map(raw.get_did2deck(self.db))
            self["codeck"] = self["codid"].map(raw.get_did2deck(self.db))
        elif table == "notes":
            self["nmodel"] = self["mid"].map(raw.get_mid2model(self.db))

        # Tags
        # ----

        if table == "notes":
            # Tags as list, rather than string joined by space
            self["ntags"] = self["ntags"].apply(
                lambda joined: [item for item in joined.split(" ") if item]
            )

        # Fields
        # ------

        if table == "notes":
            # Fields as list, rather than as string joined by \x1f
            self["nflds"] = self["nflds"].str.split("\x1f")
            self._fields_format = "list"

        # Drop columns
        # ------------

        drop_columns = set(self.columns) - set(_columns.our_columns[table])
        self.drop(drop_columns, axis=1, inplace=True)

        self._df_format = "ours"

    def raw(self, inplace=False, force=False):
        """ Bring a :class:`AnkiDataFrame` into the ``raw`` format (i.e. the
        exact format that Anki uses in its internal representation) .

        Args:
            inplace: If False, return new dataframe, else update old one
            force: If a previous conversion fails, :meth:`raw` will
                refuse to attempt another one by default. Use this option
                to force it to attempt in anyway.

        Returns:
            New :class:`AnkiDataFrame` if inplace==True, else None
        """
        if not inplace:
            df = self.copy(True)
            df.raw(inplace=True, force=force)
            return df

        if not force:
            self._check_df_format()
            if self._df_format == "anki":
                log.warning(
                    "Dataframe already is in Anki format. "
                    "Returning without doing anything."
                )
                return

        table = self._anki_table
        if table not in ["revs", "cards", "notes"]:
            self._invalid_table()

        self._df_format = "in_progress"

        # Note: Here we pretty much go through self.normalize() and revert
        # every single step.

        # Update automatic fields
        # -----------------------

        self._set_mod()
        self._set_usn()
        self._set_guid()

        # IDs
        # ---

        # Index as column:
        self.reset_index(inplace=True, drop=False)

        if table == "cards":
            self["did"] = self["cdeck"].map(raw.get_deck2did(self.db))
            self["odid"] = self["codeck"].map(raw.get_deck2did(self.db))
        if table == "notes":
            self["mid"] = self["nmodel"].map(raw.get_model2mid(self.db))

        # Fields & Hashes
        # ---------------

        if table == "notes":
            if not self._fields_format == "list":
                self.fields_as_list(inplace=True, force=True)
            # Check if success
            if not self._fields_format == "list":
                raise ValueError(
                    "It looks like the last call to fields_as_list or"
                    "fields_as_columns was not successful, so you better start "
                    "over."
                )

            # Restore the sort field.
            mids = list(self["mid"].unique())
            mid2sfld = raw.get_mid2sortfield(self.db)
            for mid in mids:
                sfield = mid2sfld[mid]
                df_model = self[self["mid"] == mid]
                fields = pd.DataFrame(df_model["nflds"].tolist())
                self.loc[self["mid"] == mid, "nsfld"] = fields[sfield].tolist()

            self["ncsum"] = self["nflds"].apply(
                lambda lst: field_checksum(lst[0])
            )

            self["nflds"] = self["nflds"].str.join("\x1f")

        # Tags
        # ----

        if table == "notes" and "nflds" in self.columns:
            self["ntags"] = self["ntags"].str.join(" ")

        # Value Maps
        # ----------

        if table in _columns.value_maps:
            for column in _columns.value_maps[table]:
                if column not in self.columns:
                    continue
                self[column] = self[column].map(
                    invert_dict(_columns.value_maps[table][column])
                )

        # Renames
        # -------

        self.rename(
            columns=invert_dict(_columns.columns_anki2ours[table]), inplace=True
        )
        self.rename(columns={"index": "id"}, inplace=True)

        # Dtypes
        # ------

        for column, typ in _columns.dtype_casts_back[table].items():
            self[column] = self[column].astype(typ)

        # Unused columns
        # --------------

        if table in ["cards", "notes"]:
            self["data"] = ""
            self["flags"] = 0

        # Drop and Rearrange
        # ------------------
        # Todo: warn about dropped columns?

        if len(self) == 0:
            new = pd.DataFrame(columns=_columns.anki_columns[table])
        else:
            new = pd.DataFrame(self[_columns.anki_columns[table]])
        self.drop(self.columns, axis=1, inplace=True)
        for col in new.columns:
            self[col] = new[col]

        self._df_format = "anki"

    # Write
    # ==========================================================================

    def summarize_changes(self, output="print") -> Optional[dict]:
        """ Summarize changes that were made with respect to the table
        as loaded from the database.

        Args:
            output: Output mode: 'print' (default: print)
                or 'dict' (return as dictionary)

        Returns:
            None or dictionary
        """
        as_dict = {
            "n": len(self),
            "n_modified": sum(self.was_modified(na=False)),
            "n_added": sum(self.was_added()),
            "n_deleted": sum(self.was_deleted()),
        }
        as_dict["has_changed"] = (
            as_dict["n_modified"] or as_dict["n_added"] or as_dict["n_deleted"]
        )
        if output == "print":
            print("Total rows: {}".format(as_dict["n"]))
            print("Compared to original version:")
            print("Modified rows: {}".format(as_dict["n_modified"]))
            print("Added rows: {}".format(as_dict["n_added"]))
            print("Deleted rows: {}".format(as_dict["n_deleted"]))
        elif output == "dict":
            return as_dict
        else:
            raise ValueError("Invalid output setting: {}".format(output))

    # Append
    # ==========================================================================

    def _get_id(self, others=()) -> int:
        """ Generate ID from timestamp and increment if it is already in use.

        .. warning::

            Do not call repeatedly without adding new IDs to index (might
            produce identical IDs). Rather use :meth:`_get_ids` instead.
        """
        idx = int(1000 * time.time())
        while idx in self.index or idx in others:
            idx += 1
        return idx

    def add_card(
        self,
        nid: int,
        cdeck: str,
        cord: Optional[Union[int, List[int]]] = None,
        cmod: Optional[int] = None,
        cusn: Optional[int] = None,
        cqueue: Optional[str] = None,
        ctype: Optional[str] = None,
        civl: Optional[int] = None,
        cfactor: Optional[int] = None,
        creps: Optional[int] = None,
        clapses: Optional[int] = None,
        cleft: Optional[int] = None,
        cdue: Optional[int] = None,
        inplace=False,
    ):
        return self.add_cards(
            nid=[nid],
            cdeck=cdeck,
            cord=cord,
            cmod=cmod,
            cusn=cusn,
            cqueue=cqueue,
            ctype=ctype,
            civl=civl,
            cfactor=cfactor,
            creps=creps,
            clapses=clapses,
            cleft=cleft,
            cdue=cdue,
            inplace=inplace,
        )

    # todo: change order of arguments?
    # fixme: cord will be replaced
    # todo: duplicate cards (same note, same cord)?
    def add_cards(
        self,
        nid: List[int],
        cdeck: Union[str, List[str]],
        cord: Optional[Union[int, List[int]]] = None,
        cmod: Optional[Union[int, List[int]]] = None,
        cusn: Optional[Union[int, List[int]]] = None,
        cqueue: Optional[Union[str, List[str]]] = None,
        ctype: Optional[Union[str, List[str]]] = None,
        civl: Optional[Union[int, List[int]]] = None,
        cfactor: Optional[Union[int, List[int]]] = None,
        creps: Optional[Union[int, List[int]]] = None,
        clapses: Optional[Union[int, List[int]]] = None,
        cleft: Optional[Union[int, List[int]]] = None,
        cdue: Optional[Union[int, List[int]]] = None,
        inplace=False,
    ):
        """
        Add cards belonging to notes of one model.

        Args:
            nid: Note IDs of the notes that you want to add cards for
            cdeck: Name of deck to add cards to as string or list of strings
                (different deck for each nid).
            cord: Number of the template to add cards for as int or list
                thereof. The template corresponds to the reviewing
                direction. If left ``None`` (default), cards for all
                templates will be added.
                It is not possible to specify different cord for different
                nids!
            cmod: List of modification timestamps.
                Will be set automatically if ``None`` (default) and it is
                discouraged to set your own.
            cusn: List of Update Sequence Numbers.
                Will be set automatically (to -1, i.e. needs update)
                if ``None`` (default) and it is
                very discouraged to set your own.
            cqueue: 'sched buried', 'user buried', 'suspended', 'new',
                'learning', 'due', 'in learning' (learning but next rev at
                least a day after the previous review). If ``None`` (default),
                'new' is chosen for all cards. Specify as string or list
                thereof.
            ctype: List of card types ('learning', 'review', 'relearn', 'cram').
                If ``None`` (default) 'learning' is chosen for all.
            civl: The new interval that the card was pushed to after the review.
                Positive values are in days, negative values are in seconds
                (for learning cards).  If ``None`` (default) 0 is chosen for
                all cards.
            cfactor: The new ease factor of the card in permille. If ``None``
                (default) 0 is chosen for all.
            creps: Number of reviews. If ``None`` (default), 0 is chosen for
                all cards.
            clapses: The number of times the card went from a 'was answered
                correctly' to 'was answered incorrectly'. If ``None`` (default),
                0 is chosen for all cards.
            cleft: Of the form ``a*1000+b``, with: ``b`` the number of reps
                left till graduation and ``a`` the number of reps left today.
                If ``None`` (default), 0 is chosen for all cards.
            cdue: Due is used differently for different card types: new:
                note id or random int, due: integer day, relative to the
                collection's creation time, learning: integer timestamp.
                If ``None`` (default), check that we're adding a new card and
                set to note ID.
            inplace: If ``False`` (default), return a new
                :class:`~ankipandas.AnkiDataFrame`, if True, modify in place and
                return new card IDs

        Returns:
            :class:`~ankipandas.AnkiDataFrame` if ``inplace==True``, else
            list of new card IDs

        """
        self._check_our_format()
        if not self._anki_table == "cards":
            raise ValueError("Cards can only be added to cards table!")

        # --- Ord ---

        nid2mid = raw.get_nid2mid(self.db)
        missing_nids = sorted(list(set(nid) - set(nid2mid.keys())))
        if missing_nids:
            raise ValueError(
                "The following note IDs (nid) can't be found in the notes "
                "table: {}. Perhaps you didn't call notes.write() to write "
                "them back into the database?".format(
                    ", ".join(map(str, missing_nids))
                )
            )

        mids = list(set(map(lambda x: nid2mid[x], nid)))
        if len(mids) >= 2:
            raise ValueError(
                "It is only supported to add cards for notes of the same model"
                ", but you're trying to add cards for notes of "
                "models: {}".format(", ".join(map(str, mids)))
            )
        mid = mids[0]

        minfo = raw.get_model_info(self.db)

        available_ords = list(map(lambda x: x["ord"], minfo[mid]["tmpls"]))
        if cord is None:
            cord = available_ords
        elif isinstance(cord, int):
            cord = [cord]
        elif is_list_like(cord):
            pass
        else:
            raise ValueError(
                "Unknown type for cord specifiation: {}".format(type(cord))
            )
        not_available = sorted(list(set(cord) - set(available_ords)))
        if not_available:
            raise ValueError(
                "The following templates are not available for notes of "
                "this model: {}".format(", ".join(map(str, not_available)))
            )

        # --- Deck ---

        if isinstance(cdeck, str):
            cdeck = [cdeck] * len(nid)
        elif is_list_like(cdeck):
            if len(cdeck) != len(nid):
                raise ValueError(
                    "Number of decks doesn't match number of "
                    "notes for which cards should be added: {} "
                    "instead of {}.".format(len(cmod), len(nid))
                )
        else:
            raise ValueError("Unknown format for cdeck: {}".format(type(cdeck)))
        unknown_decks = sorted(
            list(set(cdeck) - set(raw.get_did2deck(self.db).values()))
        )
        if unknown_decks:
            raise ValueError(
                "The following decks do not seem to exist: {}".format(
                    ", ".join(unknown_decks)
                )
            )

        # --- Rest ---

        def _handle_input(inpt, name, default, typ, options=None):
            if inpt is None:
                inpt = [default] * len(nid)
            elif is_list_like(inpt):
                if len(inpt) != len(nid):
                    raise ValueError(
                        "Number of {} doesn't match number of "
                        "notes for which cards should be added: {} "
                        "instead of {}.".format(name, len(cmod), len(nid))
                    )
            elif isinstance(inpt, typ):
                inpt = [inpt] * len(nid)
            else:
                raise ValueError(
                    "Invalid type of {} specification: {}".format(
                        name, type(inpt)
                    )
                )
            if options is not None:
                invalid = sorted(list(set(inpt) - set(options)))
                if invalid:
                    raise ValueError(
                        "The following values are no valid "
                        "entries for {}: {}".format(name, ", ".join(invalid))
                    )
            return inpt

        cmod = _handle_input(cmod, "cmod", int(time.time()), int)
        cusn = _handle_input(cusn, "cusn", -1, int)
        cqueue = _handle_input(
            cqueue,
            "cqueue",
            "new",
            str,
            options=[
                "sched buried",
                "user buried",
                "suspended",
                "new",
                "learning",
                "due",
                "in learning",
            ],
        )
        ctype = _handle_input(
            ctype,
            "ctype",
            "learning",
            str,
            options=["learning", "review", "relearn", "cram"],
        )
        civl = _handle_input(civl, "civl", 0, int)
        cfactor = _handle_input(cfactor, "cfactor", 0, int)
        creps = _handle_input(creps, "creps", 0, int)
        clapses = _handle_input(clapses, "clapses", 0, int)
        cleft = _handle_input(cleft, "cleft", 0, int)

        # --- Due ---
        # Careful: Has to come after cqueue is defined!

        if cdue is None:
            if set(cqueue) == {"new"}:
                cdue = nid
            else:
                raise ValueError(
                    "Due date can only be set automatically for cards of type"
                    "/queue 'new', but you have types: {}".format(
                        ", ".join(set(cqueue))
                    )
                )
        elif is_list_like(cdue):
            if len(cdue) != len(nid):
                raise ValueError(
                    "Number of cdue doesn't match number of "
                    "notes for which cards should be added: {} "
                    "instead of {}.".format(len(cmod), len(nid))
                )
        elif isinstance(cdue, int):
            cdue = [cdue] * len(nid)
        else:
            raise ValueError(
                "Invalid type of cdue specification: {}".format(type(cdue))
            )

        # Now we need to decide on contents for EVERY column in the DF
        all_cids = self._get_ids(n=len(nid) * len(cord))
        add = pd.DataFrame(columns=self.columns, index=all_cids)
        for icord, co in enumerate(cord):
            cid = all_cids[icord * len(nid) : (icord + 1) * len(nid)]
            known_columns = {
                "nid": nid,
                "cdeck": cdeck,
                "cord": [co] * len(nid),
                "cdue": cdue,
                "cmod": cmod,
                "cusn": cusn,
                "cqueue": cqueue,
                "ctype": ctype,
                "civl": civl,
                "cfactor": cfactor,
                "creps": creps,
                "clapses": clapses,
                "cleft": cleft,
                "codeck": [""] * len(nid),
                "codue": [0] * len(nid),
            }

            for key, item in known_columns.items():
                add.loc[cid, key] = pd.Series(item, index=cid)

        add = add.astype(
            {
                key: value
                for key, value in _columns.dtype_casts_all.items()
                if key in self.columns
            }
        )

        if not inplace:
            return self.append(add)
        else:
            replace_df_inplace(self, self.append(add))
            return all_cids

    def _get_ids(self, n=1) -> List[int]:
        """ Generate ID from timestamp and increment if it is already in use.

        Args:
            n: Number of IDs to generate
        """
        indices = []
        for i in range(n):
            indices.append(self._get_id(others=indices))
        return indices

    # Todo: If tags single list: Same for all!
    def add_notes(
        self,
        nmodel: str,
        nflds: Union[List[List[str]], Dict[str, List[str]]],
        ntags: List[List[str]] = None,
        nid=None,
        nguid=None,
        nmod=None,
        nusn=None,
        inplace=False,
    ):
        """ Add multiple new notes corresponding to one model.

        Args:
            nmodel: Name of the model (must exist already, check
                :meth:`list_models` for a list of available models)
            nflds: Fields of the note either as list of lists, e.g.
                ``[[field1_note1, ... fieldN_note1], ...,
                [field1_noteM, ... fieldN_noteM]]`` or dictionary
                ``{field name: [field_value1, ..., field_valueM]}`` or list of
                dictionaries: ``[{field_name: field_value for note 1}, ...,
                {field_name: field_value for note N}]``.
                If dictionaries are used: If fields are not present,
                they are filled with empty strings.
            ntags: Tags of the note as list of list of strings:
                ``[[tag1_note1, tag2_note1, ... ], ... [tag_1_noteM, ...]]``.
                If ``None``, no tags will be added.
            nid: List of note IDs. Will be set automatically if ``None``
                (default) and it is discouraged to set your own.
            nguid: List of Globally Unique IDs. Will be set automatically if
                ``None`` (default), and it is discouraged to set your own.
            nmod: List of modification timestamps.
                Will be set automatically if ``None`` (default) and it is
                discouraged to set your own.
            nusn: List of Update Sequence Number.
                Will be set automatically (to -1, i.e. needs update)
                if ``None`` (default) and it is
                very discouraged to set your own.
            inplace: If ``False`` (default), return a new
                :class:`~ankipandas.AnkiDataFrame`, if True, modify in place and
                return new note ID

        Returns:
            :class:`~ankipandas.AnkiDataFrame` if ``inplace==True``, else
            new note ID (int)
        """
        self._check_our_format()
        if not self._anki_table == "notes":
            raise ValueError("Notes can only be added to notes table.")

        # --- Model ---

        model2mid = raw.get_model2mid(self.db)
        if nmodel not in model2mid.keys():
            raise ValueError(
                "No model of with name '{}' exists.".format(nmodel)
            )
        field_keys = raw.get_mid2fields(self.db)[model2mid[nmodel]]

        # --- Fields ---

        if is_list_dict_like(nflds):
            n_notes = len(nflds)
            specified_fields = set(
                flatten_list_list(list(map(lambda d: list(d.keys()), nflds)))
            )
            unknown_fields = sorted(list(specified_fields - set(field_keys)))
            if unknown_fields:
                raise ValueError(
                    "Unknown fields: {}".format(", ".join(unknown_fields))
                )
            field_key2field = {
                key: list(map(lambda d: d.get(key), nflds))
                for key in field_keys
            }
        elif is_list_list_like(nflds):
            n_fields = list(set(map(len, nflds)))
            n_notes = len(nflds)
            if not (len(n_fields) == 1 and n_fields[0] == len(field_keys)):
                raise ValueError(
                    "Wrong number of items for specification of field contents:"
                    " There are {} fields for your model type, but you"
                    " specified {} items.".format(
                        len(field_keys), ", ".join(map(str, n_fields))
                    )
                )
            field_key2field = {
                field_keys[i]: list(map(lambda x: x[i], nflds))
                for i in range(len(field_keys))
            }
        elif is_dict_list_like(nflds):
            lengths = list(set(map(len, nflds.values())))
            if len(lengths) >= 2:
                raise ValueError(
                    "Inconsistent number of "
                    "fields: {}".format(", ".join(map(str, lengths)))
                )
            elif len(lengths) == 0:
                raise ValueError("Are you trying to add zero notes?")
            n_notes = lengths[0]
            field_key2field = copy.deepcopy(nflds)
            for key in field_keys:
                if key not in field_key2field:
                    field_key2field[key] = [""] * n_notes
        else:
            raise ValueError("Unsupported fields specification.")

        # --- Tags ---

        if ntags is not None:
            if len(ntags) != n_notes:
                raise ValueError(
                    "Number of tags doesn't match number of notes to"
                    " be added: {} instead of {}.".format(len(ntags), n_notes)
                )
        else:
            ntags = [[]] * n_notes

        # --- Nids ---

        if nid is not None:
            if len(nid) != n_notes:
                raise ValueError(
                    "Number of note IDs doesn't match number of notes to"
                    " be added: {} instead of {}.".format(len(nid), n_notes)
                )
        else:
            nid = self._get_ids(n=n_notes)

        already_present = sorted(list(set(nid).intersection(set(self.index))))
        if already_present:
            raise ValueError(
                "The following note IDs (nid) are "
                "already present: {}".format(", ".join(map(str, nid)))
            )

        if len(set(nid)) < len(nid):
            raise ValueError("Your note ID specification contains duplicates!")

        # --- Mod ---

        if nmod is not None:
            if len(nmod) != n_notes:
                raise ValueError(
                    "Number of modification dates doesn't match number of "
                    "notes to  be added: {} "
                    "instead of {}.".format(len(nmod), n_notes)
                )
        else:
            nmod = [int(time.time()) for _ in range(n_notes)]

        # --- Guid ---

        if nguid is not None:
            if len(nguid) != n_notes:
                raise ValueError(
                    "Number of globally unique IDs (guid) doesn't match number "
                    "of notes to  be added: {} "
                    "instead of {}.".format(len(nguid), n_notes)
                )
        else:
            nguid = [generate_guid() for _ in range(n_notes)]

        existing_guids = sorted(
            list(set(nguid).intersection(self["nguid"].unique()))
        )
        if existing_guids:
            raise ValueError(
                "The following globally unique IDs (guid) are already"
                " present: {}.".format(", ".join(map(str, existing_guids)))
            )

        # todo: make efficient
        duplicate_guids = sorted(set([g for g in nguid if nguid.count(g) >= 2]))
        if duplicate_guids:
            raise ValueError(
                "The following gloally unique IDs (guid) are not unique: ",
                ", ".join(map(str, duplicate_guids)),
            )

        # --- Usn ---

        if nusn is None:
            nusn = -1
        else:
            if len(nusn) != n_notes:
                raise ValueError(
                    "Number of update sequence numbers (usn) doesn't match"
                    "number of notes to  be added: {} "
                    "instead of {}.".format(len(nusn), n_notes)
                )

        # --- Collect all  ---

        # Now we need to decide on contents for EVERY column in the DF
        known_columns = {
            "nmodel": nmodel,
            "ntags": ntags,
            "nguid": nguid,
            "nmod": nmod,
            "nusn": nusn,
        }

        # More difficult: Field columns:
        if self._fields_format == "list":
            # Be careful with order!
            # Also need to flip dimensions
            known_columns["nflds"] = np.swapaxes(
                [field_key2field[field_key] for field_key in field_keys], 0, 1
            ).tolist()
        elif self._fields_format == "columns":
            # First we need to make sure that the df has the columns for our
            # model (perhaps this is the first note of this model that we're
            # adding, so fields_as_columns() didn't add them).
            for col in field_keys:
                if col not in self:
                    self[self.fields_as_columns_prefix + col] = ""
            # Let's first set all fields as columns to '', because we also
            # need to set those which aren't from our model:
            for col in self.columns:
                if col.startswith(self.fields_as_columns_prefix):
                    known_columns[col] = [""] * n_notes
            # Now let's fill those of our model
            for col, values in field_key2field.items():
                known_columns[self.fields_as_columns_prefix + col] = values
        else:
            raise ValueError(
                "Fields have to be in 'list' or 'columns' format, but yours "
                "are in '{}' format.".format(self._fields_format)
            )

        add = pd.DataFrame(columns=self.columns, index=nid)
        for key, item in known_columns.items():
            add.loc[:, key] = pd.Series(item, index=nid)
        add = add.astype(
            {
                key: value
                for key, value in _columns.dtype_casts_all.items()
                if key in self.columns
            }
        )
        if not inplace:
            return self.append(add)
        else:
            replace_df_inplace(self, self.append(add))
            return nid

    def add_note(
        self,
        nmodel: str,
        nflds: Union[List[str], Dict[str, str]],
        ntags=None,
        nid=None,
        nguid=None,
        nmod=None,
        nusn=-1,
        inplace=False,
    ):
        """ Add new note.

        .. note::

            For better performance, it is advisable to use :meth:`add_notes`,
            when adding many notes.

        Args:
            nmodel: Name of the model (must exist already, check
                :meth:`list_models` for a list of available models)
            nflds: Fields of the note either as list or as dictionary
                ``{field name: field value}``. In the latter case, if fields
                are not present, they are filled with empty strings.
            ntags: Tags of the note as string or Iterable thereof. Defaults to
                no tags.
            nid: Note ID. Will be set automatically by default and it is
                discouraged to set your own. If you do so and it already
                exists, the existing note will be overwritten.
            nguid: Note Globally Unique ID. Will be set automatically by
                default, and it is discouraged to set your own.
            nmod: Modification timestamp. Will be set automatically by default
                and it is discouraged to set your own.
            nusn: Update sequence number. Will be set automatically
                (to -1, i.e. needs update) if ``None`` (default) and it is
                very discouraged to set your own.
            inplace: If False (default), return a new
                :class:`ankipandas.AnkiDataFrame`, if True, modify in place and
                return new note ID

        Returns:
            :class:`ankipandas.AnkiDataFrame` if ``inplace==True``, else
            new note ID (``int``)

        """
        if is_list_like(nflds):
            nflds = [nflds]
        elif isinstance(nflds, dict):
            nflds = [nflds]
        else:
            raise ValueError(
                "Unknown type for fields specification: {}".format(type(nflds))
            )
        if ntags is not None:
            ntags = [ntags]
        if nid is not None:
            nid = [nid]
        if nguid is not None:
            nguid = [nguid]
        if nmod is not None:
            nmod = [nmod]
        if nusn is not None:
            nusn = [nusn]

        ret = self.add_notes(
            nmodel=nmodel,
            nflds=nflds,
            ntags=ntags,
            nid=nid,
            nguid=nguid,
            nmod=nmod,
            nusn=nusn,
            inplace=inplace,
        )
        if inplace:
            # We get nids back
            return ret[0]
        else:
            # We get new AnkiDataFrame back
            return ret

    # Help
    # ==========================================================================

    # todo: test?
    def help_col(self, column, ret=False) -> Union[str, None]:
        """
        Show description/help about a column. To get information about all
        columns, use the :meth:`.help_cols` method instead.

        Args:
              column: Name of the column
              ret: If True, return as string, rather than printing
        """
        df = self.help_cols(column)
        if len(df) == 0:
            raise ValueError(
                "Could not find help for your search request.".format(column)
            )
        if len(df) == 2:
            # fix for nid and cid column:
            df = self.help_cols(column, table=self._anki_table)
        if len(df) != 1:
            raise ValueError("Could not find help due to bug.")
        data = df.loc[column].to_dict()
        h = "Help for column '{}'\n".format(column)
        h += "-" * (len(h) - 1) + "\n"
        if data["Native"]:
            h += "Name in raw Anki database: " + data["AnkiColumn"] + "\n"
        h += "Information from table: " + data["Table"] + "\n"
        h += "Present by default: " + str(data["Default"]) + "\n\n"
        h += "Description: " + data["Description"]
        if ret:
            return h
        else:
            print(h)

    def help_cols(
        self, column="auto", table="all", ankicolumn="all"
    ) -> pd.DataFrame:
        """
        Show information about the columns and their interpretations. To
        get information about a single column, please use :meth:`.help_col`.

        Args:
            column: Name of a field or column (as used by us) or list thereof.
                If 'auto' (default), all columns from the current dataframe will
                be shown.
                If 'all' no filtering based on the table will be performed
            table: Possible values: 'notes', 'cards', 'revlog' or list thereof.
                If 'all' no filtering based on the table will be performed
            ankicolumn: Name of a field or column (as used by Anki) or list
                thereof.
                If 'all' no filtering based on the table will be performed

        Returns:
            Pandas DataFrame with all matches.

        .. warning::

            As there are problems with text wrapping in pandas DataFrame, this
            method might change or disappear in the future.
        """
        help_path = pathlib.Path(__file__).parent / "data" / "anki_fields.csv"
        df = pd.read_csv(help_path)
        if column == "auto":
            column = list(self.columns)
        if table is not "all":
            if isinstance(table, str):
                table = [table]
            df = df[df["Table"].isin(table)]
        if column is not "all":
            if isinstance(column, str):
                column = [column]
            df = df[df["Column"].isin(column)]
        if ankicolumn is not "all":
            if isinstance(ankicolumn, str):
                ankicolumn = [ankicolumn]
            df = df[df["AnkiColumn"].isin(ankicolumn)]
        df.set_index("Column", inplace=True)
        return df

    @staticmethod
    def help(ret=False) -> Union[str, None]:
        """ Display short help text.

        Args:
            ret: Return as string instead of printing it.

        Returns:
            string if ret==True, else None
        """
        h = (
            "This is the help for the class AnkiDataFrame, a subclass of "
            "pandas.DataFrame. \n"
            "The full documentation of all class methods "
            "unique to AnkiDataFrame can be found on "
            "https://ankipandas.readthedocs.io. \n"
            "The inherited methods from "
            "pandas.DataFrame are documented at https://pandas.pydata.org/"
            "pandas-docs/stable/reference/api/pandas.DataFrame.html.\n"
            "To get information about the fields currently in this table, "
            "please use the help_cols() method."
        )
        if ret:
            return h
        else:
            print(h)



#!/usr/bin/env python3

# std
import colorlog
import logging
from typing import Union


LOG_DEFAULT_LEVEL = logging.INFO


def get_logger():
    """ Sets up global logger. """
    _log = colorlog.getLogger("AnkiPandas")

    if _log.handlers:
        # the logger already has handlers attached to it, even though
        # we didn't add it ==> logging.get_logger got us an existing
        # logger ==> we don't need to do anything
        return _log

    _log.setLevel(LOG_DEFAULT_LEVEL)

    sh = colorlog.StreamHandler()
    log_colors = {
        "DEBUG": "cyan",
        "INFO": "green",
        "WARNING": "yellow",
        "ERROR": "red",
        "CRITICAL": "red",
    }
    formatter = colorlog.ColoredFormatter(
        "%(log_color)s%(levelname)s: %(message)s", log_colors=log_colors
    )
    sh.setFormatter(formatter)
    # Controlled by overall logger level
    sh.setLevel(logging.DEBUG)

    _log.addHandler(sh)

    return _log


def set_log_level(level: Union[str, int]) -> None:
    """ Set global log level.

    Args:
        level: Either an int
            (https://docs.python.org/3/library/logging.html#levels)
            or one of the keywords, 'critical' (only the most terrifying of log
            messages), 'error', 'warning', 'info',
            'debug' (all log messages)

    Returns:
        None
    """
    lvl = level
    if isinstance(level, str):
        lvl = getattr(logging, level.upper())
    get_logger().setLevel(lvl)


log = get_logger()



#!/usr/bin/env python3

""" These function implement the more direct interactions with the Anki
database and provide basic functionality that is then used to implement the
functionality in :class:`~ankipandas.collection.Collection`,
:class:`ankipandas.ankidf.AnkiDataFrame` etc.

.. warning::

    Please only use these function if you know what you are doing, as they
    come with less consistency checks as the functionality implemented in
    :class:`~ankipandas.collection.Collection` and
    :class:`ankipandas.ankidf.AnkiDataFrame`.
    Also note that the functions here are considered to be internal, i.e. might
    change without prior notice.
"""

# std
from collections import defaultdict
import sqlite3
import json
import pathlib
from functools import lru_cache
from typing import Dict, List, Union

# 3rd
import pandas as pd

# ours
from ankipandas.util.log import log
from ankipandas._columns import tables_ours2anki, anki_columns

CACHE_SIZE = 32


# Open/Close db
# ==============================================================================


def load_db(path: Union[str, pathlib.PurePath]) -> sqlite3.Connection:
    """
    Load database from path.

    Args:
        path: String or :class:`pathlib.PurePath`.

    Returns:
        :class:`sqlite3.Connection`
    """
    path = pathlib.Path(path)
    if not path.is_file():
        raise FileNotFoundError("Not a file/file not found: {}".format(path))
    return sqlite3.connect(str(path.resolve()))


def close_db(db: sqlite3.Connection) -> None:
    """ Close the database.

    Args:
        db:  Database (:class:`sqlite3.Connection`)

    Returns:
        None
    """
    db.close()


# Basic getters
# ==============================================================================


def get_table(db: sqlite3.Connection, table: str) -> pd.DataFrame:
    """ Get raw table from the Anki database.

    Args:
        db: Database (:class:`sqlite3.Connection`)
        table: ``cards``, ``notes`` or ``revs``

    Returns:
        :class:`pandas.DataFrame`
    """

    df = pd.read_sql_query(
        "SELECT * FROM {}".format(tables_ours2anki[table]), db
    )
    return df


def get_empty_table(table: str) -> pd.DataFrame:
    """ Get empty table

    Args:
        table: ``cards``, ``notes`` or ``revs``

    Returns:
        :class: `pandas.DataFrame`
    """
    return pd.DataFrame(columns=anki_columns[table])


@lru_cache(CACHE_SIZE)
def get_info(db: sqlite3.Connection) -> dict:
    """
    Get all other information from the databse, e.g. information about models,
    decks etc.

    Args:
        db: Database (:class:`sqlite3.Connection`)

    Returns:
        Nested dictionary.
    """
    _df = pd.read_sql_query("SELECT * FROM col ", db)
    assert len(_df) == 1
    ret = {}
    for col in _df.columns:
        val = _df[col][0]
        if isinstance(val, str):
            ret[col] = json.loads(val)
        else:
            ret[col] = val
    return ret


# Basic Setters
# ==============================================================================


def _consolidate_tables(
    df: pd.DataFrame, df_old: pd.DataFrame, mode: str, id_column="id"
):

    if not list(df.columns) == list(df_old.columns):
        raise ValueError(
            "Columns do not match: Old: {}, New: {}".format(
                ", ".join(df_old.columns), ", ".join(df.columns)
            )
        )

    old_indices = set(df_old[id_column])
    new_indices = set(df[id_column])

    # Get indices
    # -----------

    if mode == "update":
        indices = set(old_indices)
    elif mode == "append":
        indices = set(new_indices) - set(old_indices)
        if not indices:
            log.warning(
                "Was told to append to table, but there do not seem to be any"
                " new entries. "
            )
    elif mode == "replace":
        indices = set(new_indices)
    else:
        raise ValueError("Unknown mode '{}'.".format(mode))

    df = df[df[id_column].isin(indices)]

    # Apply
    # -----

    if mode == "update":
        df_new = df_old.copy()
        df_new.update(df)
    elif mode == "append":
        df_new = df_old.append(df, verify_integrity=True)
    elif mode == "replace":
        df_new = df.copy()
    else:
        raise ValueError("Unknown mode '{}'.".format(mode))

    return df_new


# fixme: update mode also can delete things if we do not have all rows
def set_table(
    db: sqlite3.Connection,
    df: pd.DataFrame,
    table: str,
    mode: str,
    id_column="id",
) -> None:
    """
    Write table back to database.

    Args:
        db: Database (:class:`sqlite3.Connection`)
        df: The :class:`pandas.DataFrame` to write
        table: Table to write to: 'notes', 'cards', 'revs'
        mode: 'update': Update all existing entries, 'append': Only append new
            entries, but do not modify, 'replace': Append, modify and delete
        id_column: Column with ID
    Returns:
        None
    """
    df_old = get_table(db, table)
    df_new = _consolidate_tables(
        df=df, df_old=df_old, mode=mode, id_column=id_column
    )
    df_new.to_sql(tables_ours2anki[table], db, if_exists="replace", index=False)


# Trivially derived getters
# ==============================================================================

# todo: Using decorators here causes the function signatures to be messed up
#  with sphinx but oh well.


@lru_cache(CACHE_SIZE)
def get_ids(db: sqlite3.Connection, table: str) -> List[int]:
    """ Get list of IDs, e.g. note IDs etc.

    Args:
        db: Database (:class:`sqlite3.Connection`)
        table: 'revs', 'cards', 'notes'

    Returns:
        Nested dictionary
    """
    return get_table(db, table)["id"].astype(int).tolist()


@lru_cache(CACHE_SIZE)
def get_deck_info(db: sqlite3.Connection) -> dict:
    """ Get information about decks.

    Args:
        db: Database (:class:`sqlite3.Connection`)

    Returns:
        Nested dictionary
    """
    return get_info(db)["decks"]


@lru_cache(CACHE_SIZE)
def get_did2deck(db: sqlite3.Connection) -> Dict[int, str]:
    """ Mapping of deck IDs (did) to deck names.

    Args:
        db: Database (:class:`sqlite3.Connection`)

    Returns:
        Dictionary mapping
    """
    dinfo = get_deck_info(db)
    _did2dec = {int(did): dinfo[did]["name"] for did in dinfo}
    return defaultdict(str, _did2dec)


@lru_cache(CACHE_SIZE)
def get_deck2did(db: sqlite3.Connection) -> Dict[str, int]:
    """ Mapping of deck names to deck IDs

    Args:
        db: Database (:class:`sqlite3.Connection`)

    Returns:
        Dictionary mapping of deck id to deck name
    """
    dinfo = get_deck_info(db)
    _did2dec = {dinfo[did]["name"]: int(did) for did in dinfo}
    return defaultdict(int, _did2dec)


@lru_cache(CACHE_SIZE)
def get_model_info(db: sqlite3.Connection) -> dict:
    """ Get information about models.

    Args:
        db: Database (:class:`sqlite3.Connection`)

    Returns:
        Nested dictionary
    """
    return {int(key): value for key, value in get_info(db)["models"].items()}


@lru_cache(CACHE_SIZE)
def get_mid2model(db: sqlite3.Connection) -> Dict[int, str]:
    """ Mapping of model IDs (mid) to model names.

    Args:
        db: Database (:class:`sqlite3.Connection`)

    Returns:
        Dictionary mapping
    """
    minfo = get_model_info(db)
    _mid2model = {int(mid): minfo[mid]["name"] for mid in minfo}
    return defaultdict(str, _mid2model)


@lru_cache(CACHE_SIZE)
def get_model2mid(db: sqlite3.Connection) -> Dict[str, int]:
    """ Mapping of model name to model ID (mid)

    Args:
        db: Database (:class:`sqlite3.Connection`)

    Returns:
        Dictionary mapping
    """
    minfo = get_model_info(db)
    _mid2model = {minfo[mid]["name"]: int(mid) for mid in minfo}
    return defaultdict(int, _mid2model)


@lru_cache(CACHE_SIZE)
def get_mid2sortfield(db: sqlite3.Connection) -> Dict[int, int]:
    """ Mapping of model ID to index of sort field. """
    minfo = get_model_info(db)
    _mid2sortfield = {mid: minfo[mid]["sortf"] for mid in minfo}
    return defaultdict(int, _mid2sortfield)


@lru_cache(CACHE_SIZE)
def get_mid2fields(db: sqlite3.Connection) -> Dict[int, List[str]]:
    """ Get mapping of model ID to field names.

    Args:
        db: Databse (:class:`sqlite3.Connection`)

    Returns:
        Dictionary mapping of model ID (mid) to list of field names.
    """
    minfo = get_model_info(db)
    return {
        int(mid): [flds["name"] for flds in minfo[mid]["flds"]] for mid in minfo
    }


@lru_cache(CACHE_SIZE)
def get_cid2nid(db: sqlite3.Connection) -> Dict[int, int]:
    """ Mapping card ID to note ID.

    Args:
        db:  Database (:class:`sqlite3.Connection`)

    Returns:
        Dictionary
    """
    cards = get_table(db, "cards")
    _cid2nid = dict(zip(cards["id"], cards["nid"]))
    return defaultdict(int, _cid2nid)


@lru_cache(CACHE_SIZE)
def get_cid2did(db: sqlite3.Connection) -> Dict[int, int]:
    """ Mapping card ID to deck ID.

    Args:
        db:  Database (:class:`sqlite3.Connection`)

    Returns:
        Dictionary
    """
    cards = get_table(db, "cards")
    _cid2did = dict(zip(cards["id"], cards["did"]))
    return defaultdict(int, _cid2did)


@lru_cache(CACHE_SIZE)
def get_nid2mid(db: sqlite3.Connection) -> Dict[int, int]:
    """ Mapping note ID to model ID.

    Args:
        db:  Database (:class:`sqlite3.Connection`)

    Returns:
        Dictionary
    """
    notes = get_table(db, "notes")
    _nid2mid = dict(zip(notes["id"], notes["mid"]))
    return defaultdict(int, _nid2mid)



#!/usr/bin/env python3

# std
import copy
from pathlib import Path

# 3rd
import pandas as pd

# todo: Docstrings, cleanup

# ours
from ankipandas.util.misc import invert_dict

tables_ours2anki = {"revs": "revlog", "cards": "cards", "notes": "notes"}
tables_anki2ours = invert_dict(tables_ours2anki)

fields_file = Path(__file__).parent / "data" / "anki_fields.csv"
fields_df = pd.read_csv(fields_file)

table2index = {"cards": "cid", "notes": "nid", "revs": "rid"}

our_tables = sorted(list(tables_ours2anki.keys()))
our_columns = {
    table: sorted(
        list(
            fields_df[(fields_df["Table"] == table) & fields_df["Default"]][
                "Column"
            ].unique()
        )
    )
    for table in our_tables
}
# Remove indices
for table, columns in our_columns.items():
    columns.remove(table2index[table])

# hard code this here, because order is important
anki_columns = {
    "cards": [
        "id",
        "nid",
        "did",
        "ord",
        "mod",
        "usn",
        "type",
        "queue",
        "due",
        "ivl",
        "factor",
        "reps",
        "lapses",
        "left",
        "odue",
        "odid",
        "flags",
        "data",
    ],
    "notes": [
        "id",
        "guid",
        "mid",
        "mod",
        "usn",
        "tags",
        "flds",
        "sfld",
        "csum",
        "flags",
        "data",
    ],
    "revs": [
        "id",
        "cid",
        "usn",
        "ease",
        "ivl",
        "lastIvl",
        "factor",
        "time",
        "type",
    ],
}

columns_ours2anki = {
    table: dict(
        zip(
            fields_df[(fields_df["Table"] == table) & fields_df["Native"]][
                "Column"
            ],
            fields_df[(fields_df["Table"] == table) & fields_df["Native"]][
                "AnkiColumn"
            ],
        )
    )
    for table in our_tables
}


columns_anki2ours = {
    table: invert_dict(columns_ours2anki[table]) for table in our_tables
}

value_maps = {
    "cards": {
        "cqueue": {
            -3: "sched buried",
            -2: "user buried",
            -1: "suspended",
            0: "new",
            1: "learning",
            2: "due",
            3: "in learning",
        },
        "ctype": {0: "learning", 1: "review", 2: "relearn", 3: "cram"},
    },
    "revs": {"rtype": {0: "learning", 1: "review", 2: "relearn", 3: "cram"}},
}

dtype_casts = {"notes": {}, "cards": {}, "revs": {}}

# todo: more precise?
dtype_casts_back = {"notes": {}, "cards": {}, "revs": {}}

# Avoiding problem with ints to floats such as here:
# https://github.com/pandas-dev/pandas/issues/4094
dtype_casts2 = {
    "cards": {
        "cord": int,
        "cmod": int,
        "cusn": int,
        "cdue": int,
        "civl": int,
        "cfactor": int,
        "creps": int,
        "clapses": int,
        "cleft": int,
        "codue": int,
    },
    "notes": {"nmod": int, "nusn": int},
    "revs": {
        "cid": int,
        "rusn": int,
        "rease": int,
        "ivl": int,
        "lastivl": int,
        "rfactor": int,
        "rtime": int,
    },
}
dtype_casts_all = copy.deepcopy(dtype_casts2["cards"])
dtype_casts_all.update(dtype_casts2["notes"])
dtype_casts_all.update(dtype_casts2["revs"])



#!/usr/bin/env python3

import random
import string

# Directly copied from anki utils!


# used in ankiweb
def _base62(num, extra=""):
    s = string
    table = s.ascii_letters + s.digits + extra
    buf = ""
    while num:
        num, i = divmod(num, len(table))
        buf = table[i] + buf
    return buf


_base91_extra_chars = "!#$%&()*+,-./:;<=>?@[]^_`{|}~"


def _base91(num):
    # all printable characters minus quotes, backslash and separators
    return _base62(num, _base91_extra_chars)


def _guid64():
    """Return a base91-encoded 64bit random number."""
    return _base91(random.randint(0, 2 ** 64 - 1))


def guid():
    """ Return globally unique ID """
    return _guid64()



#!/usr/bin/env python3

# std
import unittest

# ours
from ankipandas.util.misc import invert_dict


class TestInvertDict(unittest.TestCase):
    def test_ok(self):
        a = {1: 2, 3: 4, 5: 6}
        self.assertDictEqual(invert_dict(a), {2: 1, 4: 3, 6: 5})

    def test_fails(self):
        a = {1: 2, 3: 2}
        with self.assertRaises(ValueError):
            invert_dict(a)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

# std
import pathlib
import unittest
import tempfile
import shutil

# ours
from ankipandas.collection import Collection


class TestCollection(unittest.TestCase):
    def setUp(self):
        self.db_path = (
            pathlib.Path(__file__).parent
            / "data"
            / "few_basic_cards"
            / "collection.anki2"
        )

        self.col = Collection(self.db_path)
        self.notes = self.col.notes
        self.cards = self.col.cards
        self.revs = self.col.revs

    # Summarize changes
    # ==========================================================================

    def test_summarize_changes_uninitialized(self):
        col = Collection(self.db_path)
        sc = col.summarize_changes(output="dict")
        self.assertEqual(len(sc), 0)

    def test_summarize_changes_no_changes(self):
        self.col.summarize_changes()
        sc = self.col.summarize_changes(output="dict")
        for item in ["cards", "revs", "notes"]:
            with self.subTest(item=item):
                self.assertEqual(sc[item]["n_modified"], 0)
                self.assertEqual(sc[item]["n_added"], 0)
                self.assertEqual(sc[item]["n_deleted"], 0)
                self.assertEqual(sc[item]["has_changed"], False)

    def test_summarize_notes_changed(self):
        col = Collection(self.db_path)
        col.notes.add_tag("this_will_be_modified", inplace=True)
        sc = col.summarize_changes(output="dict")
        self.assertEqual(sc["notes"]["n_modified"], sc["notes"]["n"])

    # Writing
    # ==========================================================================

    def test_read_write_identical_trivial(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            db_path = shutil.copy2(str(self.db_path), tmpdir)
            (pathlib.Path(tmpdir) / "backups").mkdir()
            col = Collection(db_path)
            col.write(modify=True, delete=True, add=True)
            col_rel = Collection(db_path)
            self.assertTrue(col.notes.equals(col_rel.notes))
            self.assertTrue(col.cards.equals(col_rel.cards))
            self.assertTrue(col.revs.equals(col_rel.revs))

    def test_write_raises_delete(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            db_path = shutil.copy2(str(self.db_path), tmpdir)
            (pathlib.Path(tmpdir) / "backups").mkdir()
            col = Collection(db_path)
            col.notes.drop(col.notes.index, inplace=True)
            cases = [
                dict(modify=False, add=True),
                dict(modify=True, add=False),
                dict(modify=True, add=True),
            ]
            for case in cases:
                with self.subTest(**case):
                    with self.assertRaises(ValueError) as context:
                        col.write(**case, delete=False)
                    self.assertTrue(
                        "would be deleted" in str(context.exception)
                    )

    def test_write_raises_modified(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            db_path = shutil.copy2(str(self.db_path), tmpdir)
            (pathlib.Path(tmpdir) / "backups").mkdir()
            col = Collection(db_path)
            col.notes.add_tag("test", inplace=True)
            cases = [
                dict(add=False, delete=True),
                dict(add=True, delete=False),
                dict(add=True, delete=True),
            ]
            for case in cases:
                with self.subTest(**case):
                    with self.assertRaises(ValueError) as context:
                        col.write(**case, modify=False)
                    self.assertTrue(
                        "would be modified" in str(context.exception)
                    )

    def test_write_raises_added(self):
        with tempfile.TemporaryDirectory() as tmpdir:
            db_path = shutil.copy2(str(self.db_path), tmpdir)
            (pathlib.Path(tmpdir) / "backups").mkdir()
            col = Collection(db_path)
            col.notes.add_note("Basic", ["test", "back"], inplace=True)
            cases = [
                dict(modify=False, delete=True),
                dict(modify=True, delete=False),
                dict(modify=True, delete=True),
            ]
            for case in cases:
                with self.subTest(**case):
                    with self.assertRaises(ValueError) as context:
                        col.write(**case, add=False)
                    self.assertTrue(
                        "would be modified" in str(context.exception)
                    )


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env ptyhon3


def is_list_like(obj):
    """ True if object type is similar to list, tuple etc. """
    return isinstance(obj, (tuple, list))


def is_list_list_like(obj):
    """ True if object is like-like object of list-like objects """
    return is_list_like(obj) and all(map(is_list_like, obj))


def is_list_dict_like(obj):
    """ True if object is list-like object of dictionaries. """
    return is_list_like(obj) and all(map(lambda x: isinstance(x, dict), obj))


def is_dict_list_like(obj):
    """ True if object is dictionary with list-like objects as values. """
    return isinstance(obj, dict) and all(map(is_list_like, obj.values()))



#!/usr/bin/env python3

""" Convenience functions to find the database and other system locations
without the user having to specify full paths.
"""

# std
import os
import collections
import datetime
import pathlib
from functools import lru_cache
import shutil
from typing import Union

# ours
from ankipandas.util.log import log


@lru_cache(32)
def _find_db(
    search_path,
    maxdepth=6,
    filename="collection.anki2",
    break_on_first=False,
    user=None,
):
    """
    Like find_database but only for one search path at a time. Also doesn't
    raise any error, even if the search path doesn't exist.

    Returns:
        collection.defaultdict({user: [list of results]})
    """
    search_path = pathlib.Path(search_path)
    if not search_path.exists():
        log.debug(
            "_find_db: Search path '{}' does not "
            "exist.".format(str(search_path))
        )
        return collections.defaultdict(list)
    if search_path.is_file():
        if search_path.name == filename:
            return collections.defaultdict(
                list, {search_path.parent.name: [search_path]}
            )
        else:
            log.warning(
                "_find_db: Search path '{}' is a file, but filename does not "
                "match that of '{}'.".format(str(search_path), filename)
            )
            return collections.defaultdict(list)
    found = collections.defaultdict(list)
    for root, dirs, files in os.walk(str(search_path)):
        if filename in files:
            _user = os.path.basename(root)
            if user and not _user == user:
                continue
            found[_user].append(pathlib.Path(root) / filename)
            if break_on_first:
                log.debug("_find_db: Breaking after first hit.")
                break
        if maxdepth and root.count(os.sep) >= maxdepth:
            log.debug(
                "_find_db: Abort search at '{}'. "
                "Max depth exceeded.".format(str(root))
            )
            del dirs[:]
    return found


@lru_cache(32)
def find_db(
    search_paths=None,
    maxdepth=8,
    filename="collection.anki2",
    user=None,
    break_on_first=True,
) -> pathlib.Path:
    """
    Find path to anki2 database.

    Args:
        search_paths: Search path as string or pathlib object or list/iterable
            thereof. If None, some search paths are set by default.
        maxdepth: Maximal search depth.
        filename: Filename of the collection (default: ``collections.anki2``)
        user: Username to which the collection belongs. If None, search for
            databases of any user.
        break_on_first: Stop searching once a database is found. This is
            obviously faster, but you will not get any errors if there are
            multiple databases matching your criteria.

    Raises:
        If none ore more than one result is found: :class:`ValueError`

    Returns:
        pathlib.Path to the anki2 database
    """
    if not search_paths:
        log.info(
            "Searching for database. This might take some time. "
            "You can speed this up by specifying a search path or "
            "directly entering the path to your database."
        )
        # todo: Windows paths?
        search_paths = [
            "~/.local/share/Anki2/",
            "~/Documents/Anki2",
            "~/Anki2/",
            pathlib.Path.home(),
        ]
    if break_on_first:
        log.warning(
            "The search will stop at the first hit, so please verify that "
            "the result is correct (for example in case there might be more"
            "than one Anki installations)"
        )
    if isinstance(search_paths, (str, pathlib.PurePath)):
        search_paths = [search_paths]
    found = {}
    for search_path in search_paths:
        found = {
            **found,
            **_find_db(
                search_path,
                maxdepth=maxdepth,
                filename=filename,
                user=user,
                break_on_first=break_on_first,
            ),
        }
        if break_on_first:
            if user is not None:
                if user in found:
                    break
            else:
                if found:
                    break

    if user:
        if user not in found:
            raise ValueError(
                "Could not find database belonging to user {}".format(user)
            )
        found = found[user]
    else:
        if len(found.keys()) >= 2:
            raise ValueError(
                "Found databases for more than one user: {}. Please specify "
                "the user.".format(", ".join(found.keys()))
            )
        elif len(found.keys()) == 0:
            raise ValueError(
                "No database found. You might increase the search depth or "
                "specify search paths to find more."
            )
        else:
            found = list(found.values())[0]
    if len(found) >= 2:
        raise ValueError(
            "Found more than one database belonging to user {} at {}".format(
                user, ", ".join(map(str, found))
            )
        )
    found = found[0]
    log.debug("Database found at '{}'.".format(found))
    return found


@lru_cache(32)
def db_path_input(
    path: Union[str, pathlib.PurePath] = None, user: str = None
) -> pathlib.Path:
    """ Helper function to interpret user input of path to database.

    1. If no path is given, we search through some default locations
    2. If path points to a file: Take that file
    3. If path points to a directory: Search in that directory

    Args:
        path: Path to database or search path or None
        user: User name of anki collection or None

    Returns:
        Path to anki database as :class:`pathlib.Path` object

    Raises:
        If path does not exist: :class:`FileNotFoundError`
        In various other cases: :class:`ValueError`
    """
    if path is None:
        result = find_db(user=user)
    else:
        path = pathlib.Path(path)
        if not path.exists():
            raise FileNotFoundError(
                "db_path_input: File '{}' does not exist.".format(str(path))
            )
        if path.is_file():
            log.debug(
                "db_path_input: Database explicitly set to '{}'.".format(path)
            )
            result = path
        else:
            result = find_db(
                search_paths=(path,), user=user, break_on_first=False
            )
            log.info("Database found at '{}'.".format(result))
    if result:
        return result
    else:
        raise ValueError("Database could not be found.")


def db_backup_file_name() -> str:
    """ Time based file name of the backup file. """
    return "backup-ankipandas-{}.anki2".format(
        datetime.datetime.now().strftime("%Y-%m-%d-%H.%M.%S.%f")
    )


def get_anki_backup_folder(
    path: Union[str, pathlib.PurePath], nexist="raise"
) -> pathlib.Path:
    """ Return path to Anki backup folder.

    Args:
        path: Path to Aki database as :class:`pathlib.Path`
        nexist: What to do if backup folder doesn't seem to exist: ``raise`` or
            ``ignore``.

    Returns:
        Path to Anki backup folder as :class:`pathlib.Path`.
    """
    path = pathlib.Path(path)
    if not path.is_file():
        raise FileNotFoundError(
            "Database path {} seems to be invalid.".format(path)
        )
    backup_folder = path.parent / "backups"
    if nexist == "raise" and not backup_folder.is_dir():
        raise ValueError(
            "Anki backup folder corresponding to database at {} doesn't seem"
            " to exist. Perhaps you can specify a custom back "
            "folder?".format(path)
        )
    return backup_folder


def backup_db(
    db_path: Union[str, pathlib.PurePath],
    backup_folder: Union[str, pathlib.PurePath] = None,
) -> pathlib.Path:
    """
    Back up database file.

    Args:
        db_path: Path to database
        backup_folder: Path to backup folder. If None is given, the backup is
            created in the Anki backup directory.

    Returns:
        Path to newly created backup file as :class:`pathlib.Path`.
    """
    db_path = pathlib.Path(db_path)
    if backup_folder:
        backup_folder = pathlib.Path(backup_folder)
        if not backup_folder.is_dir():
            log.debug("Creating backup directory {}".format(str(backup_folder)))
            backup_folder.mkdir(parents=True)
    else:
        backup_folder = get_anki_backup_folder(db_path, nexist="raise")
    if not db_path.is_file():
        raise FileNotFoundError("Database does not seem to exist.")
    backup_path = backup_folder / db_backup_file_name()
    shutil.copy2(str(db_path), str(backup_path))
    return backup_path



#!/usr/bin/env python3

# std
import copy
import unittest
import shutil
import tempfile

# ours
from ankipandas.raw import *
from ankipandas.ankidf import AnkiDataFrame as AnkiDF
from ankipandas._columns import our_columns
from ankipandas.util.dataframe import merge_dfs


class TestRawRead(unittest.TestCase):
    def setUp(self):
        self.db_path = (
            pathlib.Path(__file__).parent
            / "data"
            / "few_basic_cards"
            / "collection.anki2"
        )
        self.db = load_db(self.db_path)

    def tearDown(self):
        close_db(self.db)

    def test_get_deck_info(self):
        info = get_deck_info(self.db)
        self.assertGreaterEqual(len(info), 2)
        self.assertIsInstance(info, dict)

    def test_get_deck_names(self):
        names = get_did2deck(self.db)
        self.assertTrue(
            set(names.values()).issuperset({"Testing", "EnglishGerman"})
        )

    def test_get_model_info(self):
        minfo = get_model_info(self.db)
        self.assertIsInstance(minfo, dict)
        self.assertGreaterEqual(len(minfo), 2)

    def test_get_model_names(self):
        names = get_mid2model(self.db)
        self.assertIn("Basic", names.values())
        self.assertIn("Cloze", names.values())
        self.assertEqual(len(names), 5)

    def test_get_field_names(self):
        fnames = get_mid2fields(self.db)
        models = get_mid2model(self.db)
        fnames = {models[mid]: fnames[mid] for mid in models}
        self.assertEqual(len(fnames), len(get_mid2model(self.db)))
        self.assertListEqual(fnames["Basic"], ["Front", "Back"])


class TestRawWrite(unittest.TestCase):
    def setUp(self):
        self.db_read_path = (
            pathlib.Path(__file__).parent
            / "data"
            / "few_basic_cards"
            / "collection.anki2"
        )
        self.db_read = load_db(self.db_read_path)
        self.db_write_dir = tempfile.TemporaryDirectory()
        self.db_write_path = (
            pathlib.Path(self.db_write_dir.name) / "collection.anki2"
        )
        self._reset()

    def _reset(self):
        shutil.copy(str(self.db_read_path), str(self.db_write_path))
        self.db_write = load_db(self.db_write_path)

    def tearDown(self):
        self.db_read.close()
        self.db_write.close()
        self.db_write_dir.cleanup()

    def _check_db_equal(self):
        notes = get_table(self.db_read, "notes")
        cards = get_table(self.db_read, "cards")
        revlog = get_table(self.db_read, "revs")
        notes2 = get_table(self.db_write, "notes")
        cards2 = get_table(self.db_write, "cards")
        revlog2 = get_table(self.db_write, "revs")
        # noinspection PyUnresolvedReferences
        self.assertListEqual(notes.values.tolist(), notes2.values.tolist())
        # noinspection PyUnresolvedReferences
        self.assertListEqual(cards.values.tolist(), cards2.values.tolist())
        # noinspection PyUnresolvedReferences
        self.assertListEqual(revlog.values.tolist(), revlog2.values.tolist())

    def test_rw_identical(self):
        notes = get_table(self.db_read, "notes")
        cards = get_table(self.db_read, "cards")
        revlog = get_table(self.db_read, "revs")
        for mode in ["update", "replace", "append"]:
            with self.subTest(mode=mode):
                self._reset()
                set_table(self.db_write, notes, "notes", mode)
                set_table(self.db_write, cards, "cards", mode)
                set_table(self.db_write, revlog, "revs", mode)
                self._check_db_equal()

    def test_update(self):
        notes2 = get_table(self.db_read, "notes")
        notes = get_table(self.db_read, "notes")
        for mode in ["update", "replace", "append"]:
            with self.subTest(mode=mode):
                self._reset()
                notes2.loc[
                    notes2["id"] == 1555579337683, "tags"
                ] = "definitelynew!"
                set_table(self.db_write, notes2, "notes", mode)
                if mode == "append":
                    self._check_db_equal()
                else:
                    notes2r = get_table(self.db_write, "notes")
                    chtag = notes2r.loc[notes2r["id"] == 1555579337683, "tags"]
                    self.assertListEqual(
                        list(chtag.values.tolist()), ["definitelynew!"]
                    )
                    unchanged = notes.loc[notes["id"] != 1555579337683, :]
                    unchanged2 = notes2r.loc[notes2["id"] != 1555579337683, :]

                    self.assertListEqual(
                        list(unchanged.values.tolist()),
                        list(unchanged2.values.tolist()),
                    )

    def test_update_append_does_not_delete(self):
        notes = get_table(self.db_read, "notes")
        cards = get_table(self.db_read, "cards")
        revs = get_table(self.db_read, "revs")
        notes.drop(notes.index)
        cards.drop(cards.index)
        revs.drop(revs.index)
        for mode in ["update", "append"]:
            with self.subTest(mode=mode):
                self._reset()
                set_table(self.db_write, notes, "notes", mode)
                set_table(self.db_write, cards, "cards", mode)
                set_table(self.db_write, revs, "revs", mode)
                self._check_db_equal()

    def test_replace_deletes(self):
        notes = get_table(self.db_read, "notes")
        cards = get_table(self.db_read, "cards")
        revs = get_table(self.db_read, "revs")
        notes = notes.drop(notes.index)
        cards = cards.drop(cards.index)
        revs = revs.drop(revs.index)
        self._reset()
        set_table(self.db_write, notes, "notes", "replace")
        set_table(self.db_write, cards, "cards", "replace")
        set_table(self.db_write, revs, "revs", "replace")
        notes = get_table(self.db_write, "notes")
        cards = get_table(self.db_write, "cards")
        revs = get_table(self.db_write, "revs")
        self.assertEqual(len(notes), 0)
        self.assertEqual(len(revs), 0)
        self.assertEqual(len(cards), 0)


class TestMergeDfs(unittest.TestCase):
    def setUp(self):
        self.df = pd.DataFrame(
            {"id_df": [1, 2, 3, 1, 1], "clash": ["a", "b", "c", "a", "a"]}
        )
        self.df_add = pd.DataFrame(
            {
                "id_add": [1, 2, 3],
                "value": [4, 5, 6],
                "drop": [7, 8, 9],
                "ignore": [10, 11, 12],
                "clash": [1, 1, 1],
            }
        )

    def test_merge_dfs(self):
        df_merged = merge_dfs(
            self.df,
            self.df_add,
            id_df="id_df",
            id_add="id_add",
            prepend="_",
            columns=["value", "drop", "clash"],
            drop_columns=["id_add", "drop"],
        )
        self.assertListEqual(
            sorted(list(df_merged.columns)),
            ["_clash", "clash", "id_df", "value"],
        )
        self.assertListEqual(sorted(list(df_merged["value"])), [4, 4, 4, 5, 6])

    def test_merge_dfs_prepend_all(self):
        df_merged = merge_dfs(
            self.df,
            self.df_add,
            id_df="id_df",
            id_add="id_add",
            prepend="_",
            prepend_clash_only=False,
        )
        self.assertListEqual(
            sorted(list(df_merged.columns)),
            ["_clash", "_drop", "_ignore", "_value", "clash", "id_df"],
        )

    def test_merge_dfs_inplace(self):
        df = copy.deepcopy(self.df)
        merge_dfs(df, self.df_add, id_df="id_df", id_add="id_add", inplace=True)
        self.assertListEqual(
            sorted(list(df.columns)),
            ["clash_x", "clash_y", "drop", "id_df", "ignore", "value"],
        )
        self.assertListEqual(sorted(list(df["value"])), [4, 4, 4, 5, 6])


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python3

# std
from typing import List, Any


def invert_dict(dct: dict) -> dict:
    """ Invert dictionary, i.e. reverse keys and values.

    Args:
        dct: Dictionary

    Returns:
        Dictionary with reversed keys and values.

    Raises:
        :class:`ValueError` if values are not unique.
    """
    if not len(set(dct.values())) == len(dct.values()):
        print(dct)
        print(sorted(list(dct.values())))
        raise ValueError("Dictionary does not seem to be invertible.")
    return {value: key for key, value in dct.items()}


def flatten_list_list(lst: List[List[Any]]) -> List[Any]:
    """ Takes a list of lists and returns a list of all elements.

    Args:
        lst: List of Lists

    Returns:
        list
    """
    return [item for sublist in lst for item in sublist]



#!/usr/bin/env python3

""" Most of the functionality of the AnkiDataFrames is already tested in
test_core, because that saves to write a lot of duplicate code.

Everything else is tested here.
"""

# std
import pathlib
import unittest
import copy

# 3rd
import numpy as np

# ours
from ankipandas.ankidf import AnkiDataFrame as AnkiDF
from ankipandas._columns import our_columns
import ankipandas.raw as raw
from ankipandas.collection import Collection
import ankipandas._columns as _columns


class TestAnkiDF(unittest.TestCase):
    def setUp(self):
        self.db_path = (
            pathlib.Path(__file__).parent
            / "data"
            / "few_basic_cards"
            / "collection.anki2"
        )
        self.db = raw.load_db(self.db_path)

        self.col = Collection(self.db_path)

        # Do not modify this one!
        self.notes = self.col.notes
        self.cards = self.col.cards
        self.revs = self.col.revs
        self.table2adf = {
            "notes": self.notes,
            "cards": self.cards,
            "revs": self.revs,
        }
        self.adfs = [self.notes, self.cards, self.revs]

        self.empty_notes = self.col.empty_notes()
        self.empty_cards = self.col.empty_cards()
        self.empty_revs = self.col.empty_revs()

    def nnotes(self):
        return self.notes.copy()

    def ncards(self):
        return self.cards.copy()

    def nrevs(self):
        return self.revs.copy()

    def nenotes(self):
        return self.empty_notes.copy()

    def necards(self):
        return self.empty_cards.copy()

    def nerevs(self):
        return self.empty_revs.copy()

    def ntable(self, table):
        if table == "notes":
            return self.nnotes()
        elif table == "cards":
            return self.ncards()
        elif table == "revs":
            return self.nrevs()
        else:
            raise ValueError("Unknown table")

    def ntable2adf(self):
        return {
            "notes": self.nnotes(),
            "cards": self.ncards(),
            "revs": self.nrevs(),
        }

    def nadfs(self):
        return [self.nnotes(), self.ncards(), self.nrevs()]

    # Test constructors
    # ==========================================================================

    def test_empty(self):
        eadfs = {
            "notes": self.col.empty_notes(),
            "cards": self.col.empty_cards(),
            "revs": self.col.empty_revs(),
        }
        for table, eadf in eadfs.items():
            self.assertEqual(len(eadf), 0)
            adf = self.table2adf[table]
            self.assertListEqual(
                sorted(list(adf.columns)), sorted(list(eadf.columns))
            )

    def test_tags(self):
        self.assertListEqual(
            list(self.notes.query("index=='1555579337683'")["ntags"].values)[0],
            ["other_test_tag"],
        )
        self.assertListEqual(
            list(self.notes.query("index=='1555579352896'")["ntags"].values)[0],
            ["some_test_tag"],
        )

    def test_cards(self):
        cards = self.cards
        self.assertGreater(len(cards), 11)
        self.assertEqual(
            list(sorted(cards.columns)), sorted(our_columns["cards"])
        )

    def test_notes(self):
        notes = self.notes
        self.assertGreater(len(notes), 6)
        self.assertEqual(
            list(sorted(notes.columns)), sorted(our_columns["notes"])
        )

    def test_get_revs(self):
        revs = self.revs
        self.assertEqual(
            list(sorted(revs.columns)), sorted(our_columns["revs"])
        )
        self.assertGreater(len(revs), 4)

    # Test merging
    # ==========================================================================

    def test_merge_notes_cards(self):
        merged = self.ncards().merge_notes()
        self.assertListEqual(
            sorted(list(merged.columns)),
            sorted(list(set(our_columns["cards"]) | set(our_columns["notes"]))),
        )

    def test_merge_notes_revs(self):
        merged = self.nrevs().merge_notes()
        self.assertListEqual(
            sorted(list(merged.columns)),
            sorted(
                list(
                    # Note: 'nid' is not a notes column.
                    set(our_columns["revs"])
                    | set(our_columns["notes"])
                    | {"nid"}
                )
            ),
        )

    def test_merge_notes_raises(self):
        with self.assertRaises(ValueError):
            self.nnotes().merge_notes()

    def test_merge_cards(self):
        merged = self.nrevs().merge_cards()
        self.assertListEqual(
            sorted(list(merged.columns)),
            sorted(list(set(our_columns["revs"]) | set(our_columns["cards"]))),
        )

    def test_merge_cards_raises(self):
        with self.assertRaises(ValueError):
            self.ncards().merge_cards()
        with self.assertRaises(ValueError):
            self.nnotes().merge_cards()

    # Test properties
    # ==========================================================================

    # Trivial cases
    # --------------------------------------------------------------------------

    def test_nids_notes(self):
        self.assertListEqual(list(self.notes.index), list(self.notes.nid))
        self.assertListEqual(
            list(self.notes.index),
            list(raw.get_table(self.db, "notes")["id"].unique()),
        )
        self.assertEqual(len(self.notes.nid.unique()), len(self.notes.nid))

    def test_cids_cards(self):
        self.assertListEqual(list(self.cards.index), list(self.cards.cid))
        self.assertListEqual(
            list(self.cards.index),
            list(raw.get_table(self.db, "cards")["id"].unique()),
        )
        self.assertEqual(len(self.cards.cid.unique()), len(self.cards.cid))

    def test_rids_revs(self):
        self.assertListEqual(list(self.revs.index), list(self.revs.rid))
        self.assertListEqual(
            list(self.revs.index),
            list(raw.get_table(self.db, "revs")["id"].unique()),
        )
        self.assertEqual(len(self.revs.rid.unique()), len(self.revs.rid))

    # Slightly more elaborate cases
    # --------------------------------------------------------------------------

    def test_nids_cards(self):
        self.assertListEqual(
            sorted(list(self.cards.nid.unique())),
            sorted(list(self.notes.nid.unique())),
        )

    def test_nids_revs(self):
        self.assertTrue(
            set(self.notes.nid.unique()).issuperset(set(self.revs.nid.unique()))
        )

    def test_nids_nexist(self):
        nexist = AnkiDF()
        with self.assertRaises(ValueError):
            # noinspection PyStatementEffect
            nexist.nid

    def test_cids_revs(self):
        self.assertTrue(
            set(self.revs.cid.unique()).issubset(set(self.cards.cid.unique()))
        )

    def test_cids_notes(self):
        with self.assertRaises(ValueError):
            # noinspection PyStatementEffect
            self.notes.cid

    # --------------------------------------------------------------------------

    def test_mids(self):
        mids2s = {
            "notes": set(self.notes.mid),
            "cards": set(self.cards.mid),
            "revs": set(self.revs.mid),
        }
        mids = set(raw.get_mid2model(self.db).keys())
        for table, mids2 in mids2s.items():
            with self.subTest(table=table):
                self.assertTrue(mids2.issubset(mids))

    def test_dids(self):
        did2s = {"cards": set(self.cards.did), "revs": set(self.revs.did)}
        dids = set(raw.get_did2deck(self.db).keys())
        for table, dids2 in did2s.items():
            with self.subTest(table=table):
                self.assertTrue(dids2.issubset(dids))

    # ==========================================================================

    def test_fields_as_columns(self):
        notes = self.nnotes()
        notes = notes.fields_as_columns()
        cols = our_columns["notes"].copy()
        cols.remove("nflds")
        prefix = notes.fields_as_columns_prefix
        new_cols = [prefix + item for item in ["Front", "Back"]]
        self.assertEqual(sorted(list(notes.columns)), sorted(cols + new_cols))

    def test_fields_as_list(self):
        # Add fields as column, remove original 'flds' column, then
        # add it back from the field columns and see if things still check
        # out
        notes = self.nnotes()
        flds = copy.copy(notes["nflds"].values)
        notes = notes.fields_as_columns().fields_as_list()
        self.assertEqual(list(flds), list(notes["nflds"].values))
        self.assertListEqual(
            sorted(list(notes.columns)), sorted(our_columns["notes"])
        )

    # Convenience
    # ==========================================================================

    def test_list_decks(self):
        decks = self.cards.list_decks()
        self.assertTrue(set(decks).issuperset({"Testing", "EnglishGerman"}))

    def test_list_models(self):
        models = self.notes.list_models()
        self.assertTrue(
            set(models).issuperset({"Basic", "Basic (and reversed card)"})
        )

    # Properties
    # ==========================================================================

    def test_prop_nid(self):
        notes, cards, revs = self.nadfs()
        with self.assertRaises(ValueError):
            notes.nid = ""
        cards.nid = "a"
        revs.nid = "a"
        # noinspection PyUnresolvedReferences
        self.assertEqual(cards.nid.unique().tolist(), ["a"])
        # noinspection PyUnresolvedReferences
        self.assertEqual(revs.nid.unique().tolist(), ["a"])

    def test_prop_cid(self):
        notes, cards, revs = self.nadfs()
        with self.assertRaises(ValueError):
            cards.cid = ""
        with self.assertRaises(ValueError):
            notes.cid = ""
        revs.cid = "a"
        self.assertEqual(revs["cid"].unique().tolist(), ["a"])

    def test_prop_rid(self):
        notes, cards, revs = self.nadfs()
        with self.assertRaises(ValueError):
            revs.rid = ""
        with self.assertRaises(ValueError):
            cards.rid = ""
        with self.assertRaises(ValueError):
            notes.rid = ""

    # Tags
    # ==========================================================================

    def test_list_tags(self):
        tags = self.notes.list_tags()
        self.assertTrue(set(tags).issuperset(["adjective", "noun"]))

    def test_remove_tags(self):
        notes = self.nnotes()
        notes = notes.remove_tag(None)
        self.assertEqual(list(set(map(tuple, notes["ntags"]))), [()])

    def test_add_tags(self):
        notes = self.nnotes().remove_tag(None).add_tag("1145")
        self.assertListEqual(list(set(map(tuple, notes["ntags"]))), [("1145",)])
        notes.add_tag("abc", inplace=True)
        self.assertListEqual(
            list(set(map(tuple, notes["ntags"]))), [("1145", "abc")]
        )
        notes.add_tag(["abc", "def"], inplace=True)
        self.assertListEqual(
            list(set(map(tuple, notes["ntags"]))), [("1145", "abc", "def")]
        )
        notes.add_tag([], inplace=True)
        self.assertListEqual(
            list(set(map(tuple, notes["ntags"]))), [("1145", "abc", "def")]
        )

    def test_has_tag(self):
        notes = self.nnotes().remove_tag(None).add_tag("1145")
        self.assertListEqual(list(notes.has_tag("1145").unique()), [True])
        self.assertListEqual(list(notes.has_tag("asdf").unique()), [False])
        self.assertListEqual(list(notes.has_tag().unique()), [True])
        self.assertListEqual(
            list(notes.has_tag(["asdf", "1145"]).unique()), [True]
        )

    def test_has_tag_natural(self):
        notes = self.notes
        self.assertListEqual(
            sorted(list(notes.has_tag(["some_test_tag"]).unique())),
            [False, True],
        )

    def test_has_tags(self):
        notes = self.nnotes().remove_tag(None).add_tag("1145")
        self.assertListEqual(list(notes.has_tags("1145").unique()), [True])
        self.assertListEqual(list(notes.has_tags("asdf").unique()), [False])
        self.assertListEqual(list(notes.has_tags().unique()), [True])
        self.assertListEqual(
            list(notes.has_tags(["asdf", "1145"]).unique()), [False]
        )
        notes = notes.add_tag("asdf")
        self.assertListEqual(
            list(notes.has_tags(["asdf", "1145"]).unique()), [True]
        )

    def test_remove_tag(self):
        notes = self.nnotes().add_tag(["1145", "asdf"])
        notes.remove_tag("1145", inplace=True)
        self.assertListEqual(list(notes.has_tag(["1145"]).unique()), [False])
        self.assertListEqual(list(notes.has_tag(["asdf"]).unique()), [True])

    # Changes
    # ==========================================================================

    def test_show_modification_unchanged(self):
        for table in ["cards", "revs", "notes"]:
            with self.subTest(table=table):
                adf = self.table2adf[table]
                self.assertEqual(np.sum(~adf.was_modified()), len(adf))
                self.assertEqual(np.sum(~adf.was_added()), len(adf))
                self.assertEqual(len(adf.was_deleted()), 0)
                self.assertEqual(np.sum(~adf.was_modified(adf)), len(adf))
                self.assertEqual(np.sum(~adf.was_added(adf)), len(adf))
                self.assertEqual(len(adf.was_deleted(adf)), 0)

    def test_show_modification_empty(self):
        for table in ["cards", "revs", "notes"]:
            with self.subTest(table=table):
                adf = self.ntable(table)
                adf["new_col"] = "blargh"
                adf_old = adf.copy()
                n = len(adf)
                adf = adf.drop(adf.index)
                self.assertEqual(np.sum(~adf.was_modified()), 0)
                self.assertEqual(np.sum(~adf.was_added()), 0)
                self.assertEqual(len(adf.was_deleted()), n)
                self.assertEqual(np.sum(~adf.was_modified(adf_old)), 0)
                self.assertEqual(np.sum(~adf.was_added(adf_old)), 0)
                self.assertEqual(len(adf.was_deleted(adf_old)), n)

    def test_show_modification_all_modified(self):
        for table in ["cards", "revs", "notes"]:
            with self.subTest(table=table):
                adf = self.ntable(table)
                adf["new_col"] = "blargh"
                adf_old = adf.copy()
                adf[adf.columns[2]] = "changed!"
                self.assertEqual(np.sum(~adf.was_modified()), 0)
                self.assertEqual(np.sum(adf.was_added()), 0)
                self.assertEqual(len(adf.was_deleted()), 0)
                # ----
                self.assertEqual(len(adf.modified_columns(only=True)), len(adf))
                self.assertEqual(
                    len(adf.modified_columns(only=False)), len(adf)
                )
                self.assertEqual(
                    list(adf.modified_columns().loc[adf.index[0]]).index(True),
                    2,
                )
                # ----
                self.assertEqual(np.sum(~adf.was_modified(adf_old)), 0)
                self.assertEqual(np.sum(adf.was_added(adf_old)), 0)
                self.assertEqual(len(adf.was_deleted(adf_old)), 0)
                # ----
                self.assertEqual(
                    len(adf.modified_columns(only=True, other=adf_old)),
                    len(adf),
                )
                self.assertEqual(
                    len(adf.modified_columns(only=False, other=adf_old)),
                    len(adf),
                )
                self.assertEqual(
                    list(
                        adf.modified_columns(other=adf_old).loc[adf.index[0]]
                    ).index(True),
                    2,
                )

    def test_show_modification_some_modified(self):
        for table in ["cards", "revs", "notes"]:
            with self.subTest(table=table):
                adf = self.ntable(table)
                adf_old = adf.copy()
                adf.loc[adf.index[0], [adf.columns[2]]] = "changed!"
                self.assertEqual(np.sum(adf.was_modified()), 1)
                self.assertEqual(adf.was_modified().tolist()[0], True)
                self.assertEqual(np.sum(adf.was_added()), 0)
                self.assertEqual(len(adf.was_deleted()), 0)
                # ----
                self.assertEqual(len(adf.modified_columns(only=True)), 1)
                self.assertEqual(
                    len(adf.modified_columns(only=False)), len(adf)
                )
                self.assertEqual(
                    list(adf.modified_columns().loc[adf.index[0]]).index(True),
                    2,
                )
                # ----
                self.assertEqual(np.sum(adf.was_modified(adf_old)), 1)
                self.assertEqual(adf.was_modified(adf_old).tolist()[0], True)
                self.assertEqual(np.sum(adf.was_added(adf_old)), 0)
                self.assertEqual(len(adf.was_deleted(adf_old)), 0)
                # ----
                self.assertEqual(
                    len(adf.modified_columns(only=True, other=adf_old)), 1
                )
                self.assertEqual(
                    len(adf.modified_columns(only=False, other=adf_old)),
                    len(adf),
                )
                self.assertEqual(
                    list(
                        adf.modified_columns(other=adf_old).loc[adf.index[0]]
                    ).index(True),
                    2,
                )

    # Formats
    # ==========================================================================

    def test_reformat_trivial(self):
        for table in ["notes", "revs", "cards"]:
            with self.subTest(table=table):
                adf = self.ntable(table)
                adf2 = adf.normalize()
                self.assertTrue(adf.equals(adf2))

    def test_convert_raw_load_raw(self):
        for table in ["notes", "revs", "cards"]:
            with self.subTest(table=table):
                adf = self.ntable(table).raw()
                df = raw.get_table(self.db, table)
                if table == "notes":
                    df["tags"] = df["tags"].str.strip()
                self.assertTrue(adf.equals(df))

    def test_raw_normalize(self):
        for table in ["notes", "revs", "cards"]:
            with self.subTest(table=table):
                adf = self.ntable(table)
                adf2 = adf.raw().normalize()
                self.assertTrue(adf.equals(adf2))

    # Update modification stamps
    # ==========================================================================

    def test_set_usn(self):
        for table in ["notes", "revs", "cards"]:
            with self.subTest(table=table):
                adf = self.ntable(table)
                print(adf.columns)
                adf[table[0] + "usn"] = 999
                adf_old = adf.copy()
                adf.loc[adf.index[0], adf_old.columns[0]] = "definitely changed"
                adf._set_usn()
                self.assertEqual(
                    adf.loc[
                        adf.index[0], _columns.columns_anki2ours[table]["usn"]
                    ],
                    -1,
                )

    def test_set_mod(self):
        for table in ["notes", "cards"]:
            with self.subTest(table=table):
                adf = self.ntable(table)
                adf_old = adf.copy()
                adf.loc[adf.index[0], adf.columns[0]] = "definitely changed"
                adf._set_mod()
                val1 = adf.loc[
                    adf.index[0], _columns.columns_anki2ours[table]["mod"]
                ]
                val_rest_1 = adf.loc[
                    adf.index[1:], _columns.columns_anki2ours[table]["mod"]
                ]
                val2 = adf_old.loc[
                    adf.index[0], _columns.columns_anki2ours[table]["mod"]
                ]
                val_rest_2 = adf.loc[
                    adf.index[1:], _columns.columns_anki2ours[table]["mod"]
                ]
                self.assertFalse(val1 == val2)
                self.assertListEqual(list(val_rest_1), list(val_rest_2))

    # New
    # ==========================================================================

    # Add cards
    # --------------------------------------------------------------------------

    @staticmethod
    def _cards_dict(card):
        return dict(
            nid=card["nid"],
            cdeck=card["cdeck"],
            cord=card["cord"],
            cmod=card["cmod"],
            cusn=card["cusn"],
            cqueue=card["cqueue"],
            ctype=card["ctype"],
            civl=card["civl"],
            cfactor=card["cfactor"],
            clapses=card["clapses"],
            cleft=card["cleft"],
            cdue=card["cdue"],
        )

    def _test_new_card_default_values(self, cards, **kwargs):
        self.assertEqual(cards["cusn"].unique().tolist(), [-1])
        self.assertEqual(cards["cqueue"].unique().tolist(), ["new"])
        self.assertEqual(cards["ctype"].unique().tolist(), ["learning"])
        self.assertEqual(cards["civl"].unique().tolist(), [0])
        self.assertEqual(cards["cfactor"].unique().tolist(), [0])
        self.assertEqual(cards["creps"].unique().tolist(), [0])
        self.assertEqual(cards["cleft"].unique().tolist(), [0])
        for key, value in kwargs.items():
            self.assertEqual(cards[key].unique().tolist(), [value])

    def test_new_cards_default_values(self):
        empty = self.necards()

        nid1 = 1555579352896
        nid2 = 1557223191575
        nids = [nid1, nid2]
        deck = list(raw.get_did2deck(self.db).values())[0]

        kwargs = dict(cdeck=deck)

        with self.subTest(type="default values single note"):
            self._test_new_card_default_values(
                empty.add_card(nid1, deck), **kwargs
            )
        with self.subTest(type="default values single card"):
            self._test_new_card_default_values(
                empty.add_card(nid1, deck, cord=0), **kwargs, cord=0
            )
        with self.subTest(type="default values several notes"):
            self._test_new_card_default_values(
                empty.add_cards(nids, deck), **kwargs
            )
        with self.subTest(type="default values several notes one cord"):
            self._test_new_card_default_values(
                empty.add_cards(nids, deck, cord=0), **kwargs, cord=0
            )

    def test_new_cards_raises_missing_nid(self):
        empty = self.necards()
        nids = [1555579352896, -15, -16]
        with self.assertRaises(ValueError) as context:
            empty.add_cards(nids, "Default")
        self.assertTrue("-15" in str(context.exception))
        self.assertTrue("-16" in str(context.exception))
        self.assertFalse("1555579352896" in str(context.exception))

    def test_new_cards_raises_inconsistent_model(self):
        empty = self.necards()
        nids = [1555579352896, 1555579337683]
        with self.assertRaises(ValueError) as context:
            empty.add_cards(nids, "Default")
        self.assertTrue("for notes of the same model" in str(context.exception))

    def test_new_cards_raises_missing_deck(self):
        empty = self.necards()
        nids = [1555579352896]
        deck = "not existing for sure"
        with self.assertRaises(ValueError) as context:
            empty.add_cards(nids, deck)
        self.assertTrue(deck in str(context.exception))

    def test_new_cards_raises_due_default_not_new(self):
        empty = self.necards()
        nids = [1555579352896]
        deck = list(raw.get_did2deck(self.db).values())[0]
        with self.assertRaises(ValueError) as context:
            empty.add_cards(nids, deck, cqueue="learning")
        self.assertTrue("Due date can only be set" in str(context.exception))

    def test_new_card_fully_specified(self):
        empty = self.necards()
        empty_combined = self.necards()

        # Careful: Need notes of the same model!
        nid1 = 1555579352896
        nid2 = 1557223191575
        deck1 = list(raw.get_did2deck(self.db).values())[0]
        deck2 = list(raw.get_did2deck(self.db).values())[1]

        init_dict1 = dict(
            nid=nid1,
            cdeck=deck1,
            cord=0,
            cmod=123,
            cusn=5,
            cqueue="learning",
            ctype="relearn",
            civl=5,
            cfactor=17,
            clapses=89,
            cleft=15,
            cdue=178,
        )
        init_dict2 = dict(
            nid=nid2,
            cdeck=deck2,
            cord=0,
            cmod=1123,
            cusn=15,
            cqueue="due",
            ctype="review",
            civl=15,
            cfactor=117,
            clapses=189,
            cleft=115,
            cdue=1178,
        )
        init_dict_combined = dict(
            nid=[nid1, nid2],
            cdeck=[deck1, deck2],
            cord=0,
            cmod=[123, 1123],
            cusn=[5, 15],
            cqueue=["learning", "due"],
            ctype=["relearn", "review"],
            civl=[5, 15],
            cfactor=[17, 117],
            clapses=[89, 189],
            cleft=[15, 115],
            cdue=[178, 1178],
        )

        cid1 = empty.add_card(**init_dict1, inplace=True)[0]
        card1 = empty.loc[cid1]
        cid2 = empty.add_card(**init_dict2, inplace=True)[0]
        card2 = empty.loc[cid2]

        cid1, cid2 = empty_combined.add_cards(
            **init_dict_combined, inplace=True
        )
        card1c = empty_combined.loc[cid1]
        card2c = empty_combined.loc[cid2]

        self.assertDictEqual(init_dict1, self._cards_dict(card1))
        self.assertDictEqual(init_dict2, self._cards_dict(card2))
        self.assertDictEqual(init_dict1, self._cards_dict(card1c))
        self.assertDictEqual(init_dict2, self._cards_dict(card2c))

        self.assertEqual(len(empty), 2)
        self.assertEqual(len(empty_combined), 2)

    def test_new_cards_vs_new_card(self):
        # Also done in test_new_card_fully_specified

        empty = self.necards()
        empty2 = self.necards()

        nid = list(raw.get_nid2mid(self.db).keys())[0]
        deck = list(raw.get_did2deck(self.db).values())[0]

        init_dict2 = dict(
            nid=[nid],
            cdeck=deck,
            cord=0,
            cmod=123,
            cusn=5,
            cqueue="learning",
            ctype="relearn",
            civl=5,
            cfactor=17,
            clapses=89,
            cleft=15,
            cdue=178,
        )
        init_dict1 = copy.deepcopy(init_dict2)
        init_dict1["nid"] = nid

        cids = empty2.add_cards(**init_dict2, inplace=True)
        card2 = empty2.loc[cids[0]]

        cid = empty.add_card(**init_dict1, inplace=True)[0]
        card1 = empty.loc[cid]

        self.assertDictEqual(self._cards_dict(card2), self._cards_dict(card1))

    # Add notes
    # --------------------------------------------------------------------------

    def test_new_notes_raises_inconsistent(self):
        with self.assertRaises(ValueError):
            self.nnotes().add_notes("Basic", [["1", "2"]], ntags=[["1"], ["2"]])
        with self.assertRaises(ValueError):
            self.nnotes().add_notes("Basic", [["1", "2"]], nid=[123, 124])
        with self.assertRaises(ValueError):
            self.nnotes().add_notes("Basic", [["1", "2"]], nguid=[123, 124])

    def test_new_notes_raises_nid_clash(self):
        with self.assertRaises(ValueError):
            self.nnotes().add_note("Basic", ["11", "12"], nid=10).add_note(
                "Basic", ["21", "22"], nid=10
            )
        with self.assertRaises(ValueError):
            self.nnotes().add_notes(
                "Basic", [["11", "12"], ["22", "22"]], nid=[10, 10]
            )

    def test_new_notes_raises_nguid_clash(self):
        with self.assertRaises(ValueError):
            self.nnotes().add_notes(
                "Basic", [["11", "12"], ["21", "22"]], nguid=[10, 10]
            )
        with self.assertRaises(ValueError):
            self.nnotes().add_note("Basic", ["11", "12"], nguid=10).add_note(
                "Basic", ["21", "22"], nguid=10
            )

    def test_new_notes_fields_as_columns(self):
        empty = self.nenotes()
        empty.add_notes(
            "Basic",
            [["field1", "field2"], ["field21", "field22"]],
            ntags=[["tag1", "tag2"], ["tag21", "tag22"]],
            nguid=["cryptic", "cryptic2"],
            nmod=[124, 1235],
            nusn=[42, 17],
            nid=[123, 125],
            inplace=True,
        )

        empty2 = self.nenotes().fields_as_columns()
        empty2.add_notes(
            "Basic",
            [["field1", "field2"], ["field21", "field22"]],
            ntags=[["tag1", "tag2"], ["tag21", "tag22"]],
            nguid=["cryptic", "cryptic2"],
            nmod=[124, 1235],
            nusn=[42, 17],
            nid=[123, 125],
            inplace=True,
        )

        self.assertDictEqual(
            empty.fields_as_columns().to_dict(), empty2.to_dict()
        )

    @staticmethod
    def _notes_dict(notes):
        return {
            "nmodel": notes["nmodel"],
            "nflds": notes["nflds"],
            "ntags": notes["ntags"],
            "nguid": notes["nguid"],
            "nmod": notes["nmod"],
            "nusn": notes["nusn"],
        }

    def test_new_note_empty_fully_specified(self):
        empty = self.nenotes()

        init_dict = dict(
            nmodel="Basic",
            nflds=["field1", "field2"],
            ntags=["tag1", "tag2"],
            nguid="cryptic",
            nmod=124,
            nusn=42,
        )
        nid = empty.add_note(nid=123, **init_dict, inplace=True)
        self.assertEqual(nid, 123)
        note = empty.loc[nid]
        self.assertDictEqual(init_dict, self._notes_dict(note))
        self.assertEqual(len(empty), 1)

        init_dict2 = dict(
            nmodel="Basic",
            nflds=["field21", "field22"],
            ntags=["tag21", "tag22"],
            nguid="cryptic2",
            nmod=1235,
            nusn=17,
        )
        nid = empty.add_note(nid=125, **init_dict2, inplace=True)
        self.assertEqual(nid, 125)
        note = empty.loc[125]
        self.assertDictEqual(init_dict2, self._notes_dict(note))
        self.assertEqual(len(empty), 2)

        empty2 = self.nenotes()
        empty2.add_notes(
            "Basic",
            [["field1", "field2"], ["field21", "field22"]],
            ntags=[["tag1", "tag2"], ["tag21", "tag22"]],
            nguid=["cryptic", "cryptic2"],
            nmod=[124, 1235],
            nusn=[42, 17],
            nid=[123, 125],
            inplace=True,
        )
        self.assertTrue(empty.equals(empty2))

    def test_new_note_raises_suplicate(self):
        empty = self.nenotes()
        empty.add_note("Basic", ["f1", "f2"], nid=10, inplace=True)
        self.assertEqual(len(empty), 1)
        with self.assertRaises(ValueError):
            empty.add_note("Basic", ["f3", "f4"], nid=10, inplace=True)

    def test_new_note_default_values(self):
        empty = self.nenotes()

        init_dict = dict(nmodel="Basic", nflds=["field1", "field2"])
        nid = empty.add_note(nid=123, **init_dict, inplace=True)
        self.assertEqual(nid, 123)
        note = empty.loc[nid].to_dict()
        self.assertEqual(note["nmodel"], init_dict["nmodel"])
        self.assertEqual(note["nflds"], init_dict["nflds"])

    def test_new_note_raises(self):
        empty = self.nenotes()
        with self.assertRaises(ValueError):
            empty.add_note("doesntexist", [])
        with self.assertRaises(ValueError):
            empty.add_note("Basic", ["1", "2", "3"])

    def test_new_notes_equivalent_field_specifications(self):
        empty1 = self.nenotes()
        empty2 = self.nenotes()
        empty3 = self.nenotes()

        empty1.add_notes("Basic", [["11", "12"], ["21", "22"]], inplace=True)
        empty2.add_notes(
            "Basic",
            [{"Front": "11", "Back": "12"}, {"Front": "21", "Back": "22"}],
            inplace=True,
        )
        empty3.add_notes(
            "Basic", {"Front": ["11", "21"], "Back": ["12", "22"]}, inplace=True
        )
        self.assertListEqual(empty1["nflds"].tolist(), empty2["nflds"].tolist())
        self.assertListEqual(empty2["nflds"].tolist(), empty3["nflds"].tolist())

    def test_new_notes_equivalent_field_specifications_fields_as_columns(self):
        empty1 = self.nenotes().fields_as_columns()
        empty2 = self.nenotes().fields_as_columns()
        empty3 = self.nenotes().fields_as_columns()

        empty1.add_notes("Basic", [["11", "12"], ["21", "22"]], inplace=True)
        empty2.add_notes(
            "Basic",
            [{"Front": "11", "Back": "12"}, {"Front": "21", "Back": "22"}],
            inplace=True,
        )
        empty3.add_notes(
            "Basic", {"Front": ["11", "21"], "Back": ["12", "22"]}, inplace=True
        )

        p = empty1.fields_as_columns_prefix

        self.assertListEqual(
            empty1[p + "Front"].tolist(), empty2[p + "Front"].tolist()
        )
        self.assertListEqual(
            empty2[p + "Front"].tolist(), empty3[p + "Front"].tolist()
        )
        self.assertListEqual(
            empty1[p + "Back"].tolist(), empty2[p + "Back"].tolist()
        )
        self.assertListEqual(
            empty2[p + "Back"].tolist(), empty3[p + "Back"].tolist()
        )

    # Help
    # ==========================================================================

    def test_help_col(self):
        for table, adf in self.table2adf.items():
            with self.subTest(table=table):
                cols = list(adf.columns) + [adf.index.name]
                for col in cols:
                    self.assertIsInstance(adf.help_col(col, ret=True), str)

    def test_help_cols_auto(self):
        for table, adf in self.table2adf.items():
            with self.subTest(table=table):
                df = adf.help_cols()
                self.assertListEqual(
                    list(df.columns),
                    ["AnkiColumn", "Table", "Description", "Native", "Default"],
                )
                self.assertListEqual(
                    sorted(adf.columns),
                    sorted(list(set(df.index))),  # nid, cid appear twice
                )

    def test_help(self):
        notes = self.notes
        hlp = notes.help(ret=True)
        self.assertTrue(isinstance(hlp, str))


if __name__ == "__main__":
    unittest.main()
